import random

from django.core.management.base import BaseCommand
from django.utils import timezone

from app.testblog.models import Category, Post


class Command(BaseCommand):
    help = 'Install/Update sample data'

    def handle(self, *args, **options):
        categories = [
            'Python3', 'Python2', 'PHP', 'Haskell', 'SQL', 'Java', 'Javascript', 'C#', 'C++', 'Ruby',
            'Basic', 'Lisp', 'Lua', 'Perl', 'Erlang', 'F#', 'C', 'Cobol', 'go', 'Rust', 'Fortran', 'Objective-C',
            'Smalltalk', 'Turing', 'Bash', 'VBScript', 'HTML', 'CSS', 'XML', 'JSON', 'Scala', 'IO'
        ]

        categories_objects = []
        for category in categories:
            cat = Category(name=category, description=category)
            cat.save()
            categories_objects.append(cat)

        for x in range(100):
            cat = random.choice(categories_objects)
            Post(
                title='Blog post {} about {}'.format(x, str(cat)),
                content='some content',
                publish_date=timezone.now(),
                category=cat,
            ).save()





"""
trionyx.views
~~~~~~~~~~~~~

:copyright: 2018 by Maikel Martens
:license: GPLv3
"""
import csv
import io
import json
import logging
from collections import OrderedDict

from django.apps import apps
from django.views.generic import (
    View,
    TemplateView,
    DetailView,
    UpdateView as
    DjangoUpdateView,
    CreateView as DjangoCreateView,
    DeleteView as DjangoDeleteView
)
from django.core.exceptions import PermissionDenied
from django.http import Http404, StreamingHttpResponse
from django_jsend import JsendView
from django.urls import reverse
from django.core.paginator import Paginator
from django.contrib import messages
from watson import search as watson
from django.contrib.contenttypes.models import ContentType
from django.utils.translation import ugettext_lazy as _

from trionyx.views.mixins import ModelClassMixin, SessionValueMixin, ModelPermissionMixin
from trionyx.forms.helper import FormHelper
from trionyx.models import filter_queryset_with_user_filters

logger = logging.getLogger(__name__)


# =============================================================================
# List view
# =============================================================================
class ListView(ModelPermissionMixin, TemplateView, ModelClassMixin):
    """List view for showing model"""

    permission_type = 'view'
    template_name = "trionyx/core/model_list.html"

    model = None
    """Model class default will try get model based on url kwargs app, model"""

    title = None
    """Title of page default is model verbose_name_plural"""

    ajax_url = None
    """Ajax url used to get model list data"""

    download_url = None
    """Download url used to get model list data"""

    def get_title(self):
        """Get page title"""
        if self.title:
            return self.title
        return self.get_model_class()._meta.verbose_name_plural

    def get_ajax_url(self):
        """Get ajax url"""
        if self.ajax_url:
            return self.ajax_url
        return reverse('trionyx:model-list-ajax', kwargs=self.kwargs)

    def get_download_url(self):
        """Get ajax url"""
        if self.download_url:
            return self.download_url
        return reverse('trionyx:model-list-download', kwargs=self.kwargs)

    def get_create_url(self):
        """Get create url"""
        return reverse('trionyx:model-create', kwargs=self.kwargs)

    def get_mass_delete_url(self):
        """Get mass delete url"""
        return reverse('trionyx:model-mass-delete', kwargs=self.kwargs)

    def get_mass_update_url(self):
        """Get mass delete url"""
        return reverse('trionyx:model-mass-update', kwargs=self.kwargs)

    def get_context_data(self, **kwargs):
        """Add context data to view"""
        context = super().get_context_data(**kwargs)
        context.update({
            'title': self.get_title(),
            'content_type_id': ContentType.objects.get_for_model(self.get_model_class()).id,
            'ajax_url': self.get_ajax_url(),
            'download_url': self.get_download_url(),
            'create_url': self.get_create_url(),
            'mass_delete_url': self.get_mass_delete_url(),
            'mass_update_url': self.get_mass_update_url(),
            'create_permission': self.request.user.has_perm('{app_label}.add_{model_name}'.format(
                app_label=self.get_model_config().app_label,
                model_name=self.get_model_config().model_name,
            ).lower()) and not self.get_model_config().disable_add,
            'change_permission': self.request.user.has_perm('{app_label}.change_{model_name}'.format(
                app_label=self.get_model_config().app_label,
                model_name=self.get_model_config().model_name,
            ).lower()) and not self.get_model_config().disable_change,
            'delete_permission': self.request.user.has_perm('{app_label}.delete_{model_name}'.format(
                app_label=self.get_model_config().app_label,
                model_name=self.get_model_config().model_name,
            ).lower()) and not self.get_model_config().disable_delete
        })
        return context


class ModelListMixin(ModelClassMixin, SessionValueMixin):
    """Mixin for handling model list"""

    def get_page(self, paginator):
        """Get current page or page in session"""
        page = int(self.get_and_save_value('page', 1))
        if page < 1:
            return self.save_value('page', 1)

        if page > paginator.num_pages:
            return self.save_value('page', paginator.num_pages)
        return page

    def get_page_size(self):
        """Get current page size or page size in session"""
        return self.get_and_save_value('page_size', '10')

    def get_sort(self):
        """Get current sort or sort in session"""
        return self.get_and_save_value('sort', self.get_model_config().list_default_sort)

    def get_search(self):
        """Get current search or search from session, reset page if search is changed"""
        old_search = self.get_session_value('search', '')
        search = self.get_and_save_value('search', '')
        if old_search != search:
            self.page = 1
            self.get_session_value('page', self.page)
        return search

    def get_filters(self):
        """Get all active filters"""
        try:
            return json.loads(self.get_and_save_value('filters', '[]'))
        except json.JSONDecodeError:
            return []

    def get_all_fields(self):
        """Get all available fields"""
        return {
            name: {
                'name': name,
                'label': field['label'],
                'type': field['type'],
                'choices': field['choices'],
            }
            for name, field in self.get_model_config().get_list_fields().items()
        }

    def get_current_fields(self):
        """Get current list to be used"""
        if hasattr(self, 'current_fields') and self.current_fields:
            return self.current_fields

        field_attribute = 'list_{}_{}_fields'.format(self.kwargs.get('app'), self.kwargs.get('model'))
        current_fields = self.request.user.get_attribute(field_attribute, [])
        request_fields = self.request.POST.get('selected_fields', None)

        if request_fields and ','.join(current_fields) != request_fields:
            # TODO validate fields
            current_fields = list(filter(lambda f: f != 'NR', request_fields.split(',')))
            self.request.user.set_attribute(field_attribute, current_fields)
        elif request_fields:
            current_fields = request_fields.split(',')

        if not current_fields:
            config = self.get_model_config()
            current_fields = config.list_default_fields if config.list_default_fields else ['id']

        # TODO check if all fields are still valid, filter out invalid and save
        self.current_fields = current_fields
        return current_fields

    def get_paginator(self):
        """Get paginator"""
        return Paginator(self.get_queryset(), self.get_page_size())

    def get_queryset(self):
        """Get qeuryset for model"""
        query = self.search_queryset()
        query = filter_queryset_with_user_filters(query, self.get_filters(), self.request)

        fields = self.get_all_fields()
        select_related = self.get_model_config().list_select_related if self.get_model_config().list_select_related else []
        for field in self.get_current_fields():
            field_parts = field.split('__')
            if len(field_parts) > 1:
                select_related.append('__'.join(field_parts[:-1]))
            elif fields[field]['type'] == 'related':
                select_related.append(field)

        if select_related:
            query = query.select_related(*set(select_related))

        return query.order_by(self.get_sort())

    def search_queryset(self):
        """Get search query set"""
        queryset = self.get_model_class().objects.get_queryset()
        return watson.filter(queryset, self.get_search(), ranking=False)


class ListJsendView(ModelPermissionMixin, JsendView, ModelListMixin):
    """Ajax list view"""

    permission_type = 'view'

    def __init__(self, *args, **kwargs):
        """Init ListJsendView"""
        super().__init__(*args, **kwargs)
        self.page = None
        self.page_size = None
        self.sort = None
        self.fields = None

    def handle_request(self, request, *args, **kwargs):
        """Give back list items + config"""
        paginator = self.get_paginator()
        # Call search first, it will reset page if search is changed
        search = self.get_search()
        page = self.get_page(paginator)
        items = self.get_items(paginator, page)
        return {
            'search': search,
            'filters': self.get_filters(),
            'page': page,
            'page_size': self.get_page_size(),
            'num_pages': paginator.num_pages,
            'count': paginator.count,
            'sort': self.get_sort(),
            'current_fields': self.get_current_fields(),
            'fields': self.get_all_fields(),

            'items': items,
        }

    def get_items(self, paginator, current_page):
        """Get list items for current page"""
        fields = self.get_model_config().get_list_fields()

        page = paginator.page(current_page)

        items = []
        for item in page:
            items.append({
                'id': item.id,
                'url': self.get_model_config().get_absolute_url(item),
                'row_data': [
                    fields[field]['renderer'](item, field)
                    for field in self.get_current_fields()
                ]
            })
        return items


class ListExportView(ModelPermissionMixin, View, ModelListMixin):
    """View for downloading an export of a list view"""

    permission_type = 'view'

    def post(self, request, app, model, **kwargs):
        """Handle post request"""
        return self.csv_response()

    def items(self):
        """Get all list items"""
        query = self.get_queryset()
        fields = self.get_model_config().get_list_fields()

        for item in query.iterator():
            row = OrderedDict()
            for field_name in self.get_current_fields():
                field = fields.get(field_name)
                if not field_name:
                    row[field_name] = ''

                if hasattr(item, field['field']):
                    row[field_name] = getattr(item, field['field'])
                else:
                    row[field_name] = ''  # TODO Maybe render field ans strip html?
            yield row

    def csv_response(self):
        """Get csv response"""
        def stream():
            """Create data stream generator"""
            stream_file = io.StringIO()
            csvwriter = csv.writer(stream_file, delimiter=',', quotechar='"')

            csvwriter.writerow(self.get_current_fields())

            for index, item in enumerate(self.items()):
                csvwriter.writerow([value for index, value in item.items()])
                stream_file.seek(0)
                data = stream_file.read()
                stream_file.seek(0)
                stream_file.truncate()
                yield data

        response = StreamingHttpResponse(stream(), content_type="text/csv")
        response["Content-Disposition"] = "attachment; filename={}.csv".format(self.get_model_config().model_name.lower())
        return response


class ListChoicesJsendView(ModelPermissionMixin, JsendView, ModelListMixin):
    """View for getting choices list for related field"""

    permission_type = 'view'

    def handle_request(self, request, *args, **kwargs):
        """Build choices list for related field"""
        try:
            ModelClass = self.get_model_class()
            fields = request.GET.get('field').split('__')
            RelatedClass = getattr(ModelClass, fields[0]).field.related_model

            for field in fields[1:]:
                RelatedClass = getattr(RelatedClass, field).field.related_model
        except Exception:
            return []
        return list(RelatedClass.objects.values_list('id', 'verbose_name'))


# =============================================================================
# Detail tab views
# =============================================================================
class DetailTabView(ModelPermissionMixin, DetailView, ModelClassMixin):
    """Detail tab view, shows model details in tab view"""

    permission_type = 'view'

    template_name = "trionyx/core/model_view.html"
    """Template name for rendering the view, default is 'trionyx/core/detail_tab_view.html'

        .. note::
            Only overide this if you want to customize the layout of the page.
    """

    model_alias = None
    """Model identifier, default is '<model app name>.<model name>'."""

    title = None
    """Page title if not set object __str__ is used"""

    def get_queryset(self):
        """Get queryset based on url params(<app>, <mode>) if model is not set on class"""
        if self.queryset is None and not self.model:
            try:
                ModelClass = apps.get_model(self.kwargs.get('app'), self.kwargs.get('model'))
                return ModelClass._default_manager.all()
            except LookupError:
                raise Http404()
        return super().get_queryset()

    def get_context_data(self, **kwargs):
        """Add context data to view"""
        context = super().get_context_data(**kwargs)
        tabs = self.get_active_tabs()
        context.update({
            'page_detail_tabs': tabs,
            'active_tab': tabs[0].code if tabs else '',
            'app_label': self.get_app_label(),
            'model_name': self.get_model_name(),
            'model_alias': self.get_model_alias(),
            'model_verbose_name': self.object._meta.verbose_name.title(),
            'view_header_buttons': list(self.view_header_buttons()),
            'back_url': self.get_back_url(),
            'edit_url': self.get_edit_url(),
            'delete_url': self.get_delete_url(),
            'title': self.title,
            'change_permission': self.request.user.has_perm('{app_label}.change_{model_name}'.format(
                app_label=self.get_model_config().app_label,
                model_name=self.get_model_config().model_name,
            ).lower()) and not self.get_model_config().disable_change,
            'delete_permission': self.request.user.has_perm('{app_label}.delete_{model_name}'.format(
                app_label=self.get_model_config().app_label,
                model_name=self.get_model_config().model_name,
            ).lower()) and not self.get_model_config().disable_delete
        })
        return context

    def get_back_url(self):
        """Get back url"""
        return reverse('trionyx:model-list', kwargs={
            'app': self.get_app_label(),
            'model': self.get_model_name(),
        })

    def get_delete_url(self):
        """Get model object delete url"""
        return reverse('trionyx:model-delete', kwargs={
            'app': self.get_app_label(),
            'model': self.get_model_name(),
            'pk': self.object.id
        })

    def get_edit_url(self):
        """Get model object edit url"""
        return reverse('trionyx:model-edit', kwargs={
            'app': self.get_app_label(),
            'model': self.get_model_name(),
            'pk': self.object.id
        })

    def view_header_buttons(self):
        """Get the view header buttons"""
        from django.urls.exceptions import NoReverseMatch

        def url_reverse(url):
            kwargs_list = [
                {
                    'app': self.get_app_label(),
                    'model': self.get_model_name(),
                    'pk': self.object.id
                },
                {
                    'pk': self.object.id
                },
                {}
            ]

            for kwargs in kwargs_list:
                try:
                    return reverse(url, kwargs=kwargs)
                except NoReverseMatch:
                    pass

            raise NoReverseMatch('Could not find match for {}'.format(url))

        if self.get_model_config().view_header_buttons:
            for config in self.get_model_config().view_header_buttons:
                if 'show' in config and not config['show'](self.object, self.get_model_alias()):
                    continue

                button_type = config.get('type', 'bg-theme')
                yield {
                    'label': config['label'](self.object, self.get_model_alias()) if callable(config['label']) else config['label'],
                    'type': button_type(self.object, self.get_model_alias()) if callable(button_type) else button_type,
                    'url': config['url'](self.object, self.get_model_alias()) if callable(config['url']) else url_reverse(config['url']),
                    'modal': config.get('modal', True)
                }

    def get_model_alias(self):
        """Get model alias"""
        if self.model_alias:
            return self.model_alias
        return '{}.{}'.format(self.get_app_label(), self.get_model_name())

    def get_app_label(self):
        """Get model app label"""
        return self.object._meta.app_label

    def get_model_name(self):
        """Get model name"""
        return self.object._meta.model_name

    def get_active_tabs(self):
        """Get all active tabs"""
        from trionyx.views import tabs
        return list(tabs.get_tabs(self.get_model_alias(), self.object))

    def dispatch(self, request, *args, **kwargs):
        """Validate if user can use view"""
        if False:  # TODO do permission check based on Model
            raise PermissionDenied
        return super().dispatch(request, *args, **kwargs)


class DetailTabJsendView(ModelPermissionMixin, JsendView, ModelClassMixin):
    """View for getting tab view with ajax"""

    permission_type = 'view'

    def handle_request(self, request, app, model, pk):
        """Render and return tab"""
        from trionyx.views import tabs
        ModelClass = self.get_model_class()
        object = ModelClass.objects.get(id=pk)

        tab_code = request.GET.get('tab')
        model_alias = request.GET.get('model_alias')
        model_alias = model_alias if model_alias else '{}.{}'.format(app, model)

        # TODO permission check

        item = tabs.get_tab(model_alias, object, tab_code)

        return item.get_layout(object).render(request)


class LayoutView(DetailTabView):
    """Display layout for model"""

    permission_type = 'view'

    template_name = "trionyx/core/layout_view.html"

    def get_context_data(self, **kwargs):
        """Add layout to context"""
        context = super().get_context_data(**kwargs)
        from trionyx.views import layouts
        try:
            context['layout'] = layouts.get_layout(
                self.kwargs.get('code'),
                self.object,
                self.request
            )
        except Exception:
            raise Http404()
        return context


# =============================================================================
# Update/Create/Delete view
# =============================================================================
class UpdateView(ModelPermissionMixin, DjangoUpdateView, ModelClassMixin):
    """Update view that renders view with crispy-forms"""

    permission_type = 'change'

    template_name = 'trionyx/core/model_update.html'

    title = None
    """Title of page, default is: Update <object.__str__>"""

    submit_value = None
    """Value of the form submit button"""

    cancel_url = None
    """Url code for cancel button, when not set object.get_absolute_url is used"""

    @property
    def success_url(self):
        """Return success url"""
        return self.get_model_config().get_absolute_url(self.object)

    def get_queryset(self):
        """Get queryset based on url params(<app>, <mode>) if model is not set on class"""
        if self.queryset is None and not self.model:
            try:
                return self.get_model_class()._default_manager.all()
            except LookupError:
                raise Http404()
        return super().get_queryset()

    def get_form_class(self):
        """Get form class for model"""
        if self.form_class:
            return self.form_class
        from trionyx.forms import form_register
        if self.kwargs.get('code'):
            return form_register.get_form(self.get_model_class(), self.kwargs.get('code'))
        return form_register.get_edit_form(self.get_model_class())

    def get_form(self, form_class=None):
        """Get form for model"""
        form = super().get_form(form_class)

        if not getattr(form, 'helper', None):
            form.helper = FormHelper(form)
            form.helper.form_tag = False
        else:
            form.helper.form_tag = False
        return form

    def get_context_data(self, **kwargs):
        """Add context data to view"""
        context = super().get_context_data(**kwargs)

        context.update({
            'title': self.title,
            'submit_value': self.submit_value,
            'cancel_url': self.cancel_url,
            'object_url': self.get_model_config().get_absolute_url(self.object),
        })

        return context

    def form_valid(self, form):
        """Add success message"""
        response = super().form_valid(form)
        messages.success(self.request, _("Successfully saved ({})").format(self.object))
        return response

    def form_invalid(self, form):
        """Form invalid"""
        logger.debug(json.dumps(form.errors))
        return super().form_invalid(form)


class CreateView(ModelPermissionMixin, DjangoCreateView, ModelClassMixin):
    """Create view that renders view with crispy-forms"""

    permission_type = 'add'

    template_name = 'trionyx/core/model_create.html'

    title = None
    """Title of page, default is: Update <object.__str__>"""

    submit_value = None
    """Value of the form submit button"""

    cancel_url = None
    """Url code for cancel button, when not set model list view is used"""

    @property
    def success_url(self):
        """Return success url"""
        return self.get_model_config().get_absolute_url(self.object)

    def get_queryset(self):
        """Get queryset based on url params(<app>, <mode>) if model is not set on class"""
        if self.queryset is None and not self.model:
            try:
                return self.get_model_class()._default_manager.all()
            except LookupError:
                raise Http404()
        return super().get_queryset()

    def get_form_class(self):
        """Get form class for model"""
        if self.form_class:
            return self.form_class
        from trionyx.forms import form_register
        if self.kwargs.get('code'):
            return form_register.get_form(self.get_model_class(), self.kwargs.get('code'))
        return form_register.get_create_form(self.get_model_class())

    def get_form(self, form_class=None):
        """Get form for model"""
        form = super().get_form(form_class)

        if not getattr(form, 'helper', None):
            form.helper = FormHelper(form)
            form.helper.form_tag = False
        else:
            form.helper.form_tag = False
        return form

    def get_cancel_url(self):
        """Get cancel url"""
        if self.cancel_url:
            return self.cancel_url

        ModelClass = self.get_model_class()
        return reverse('trionyx:model-list', kwargs={
            'app': ModelClass._meta.app_label,
            'model': ModelClass._meta.model_name,
        })

    def get_context_data(self, **kwargs):
        """Add context data to view"""
        context = super().get_context_data(**kwargs)

        context.update({
            'title': self.title,
            'submit_value': self.submit_value,
            'cancel_url': self.get_cancel_url(),
            'model_verbose_name': self.get_model_class()._meta.verbose_name,
        })

        return context

    def form_valid(self, form):
        """Add success message"""
        response = super().form_valid(form)
        messages.success(self.request, _("Successfully created ({})").format(self.object))
        return response

    def form_invalid(self, form):
        """Form invalid"""
        logger.debug(json.dumps(form.errors))
        return super().form_invalid(form)


class DeleteView(ModelPermissionMixin, DjangoDeleteView, ModelClassMixin):
    """Delete view"""

    permission_type = 'delete'

    template_name = 'trionyx/core/model_delete.html'

    title = None
    """Title of page, default is: Update <object.__str__>"""

    submit_value = None
    """Value of the form submit button"""

    cancel_url = None
    """Url code for cancel button, when not set model list view is used"""

    def get_queryset(self):
        """Get queryset based on url params(<app>, <mode>) if model is not set on class"""
        if self.queryset is None and not self.model:
            try:
                return self.get_model_class()._default_manager.all()
            except LookupError:
                raise Http404()
        return super().get_queryset()

    def get_context_data(self, **kwargs):
        """Add context data to view"""
        context = super().get_context_data(**kwargs)
        context.update({
            'title': self.title,
            'submit_value': self.submit_value,
            'cancel_url': self.cancel_url,
            'object_url': self.get_model_config().get_absolute_url(self.object)
        })
        return context

    def get_success_url(self):
        """Get success url"""
        messages.success(self.request, _("Successfully deleted ({})").format(self.object))
        if self.success_url:
            return reverse(self.success_url)

        if 'app' in self.kwargs and 'model' in self.kwargs:
            return reverse('trionyx:model-list', kwargs={
                'app': self.kwargs.get('app'),
                'model': self.kwargs.get('model'),
            })

        return '/'



"""
trionyx.views.mixins
~~~~~~~~~~~~~~~~~~~~

:copyright: 2018 by Maikel Martens
:license: GPLv3
"""
from typing import Optional

from django.apps import apps
from django.db.models import Model
from django.core.exceptions import PermissionDenied

from trionyx.config import models_config, ModelConfig


class ModelClassMixin:
    """Mixen for getting model class"""

    def get_model_class(self) -> Optional[Model]:
        """Get model class"""
        if getattr(self, 'model', None):
            return self.model
        elif getattr(self, 'object', None):
            return self.object.__class__
        elif 'app' in self.kwargs and 'model' in self.kwargs:
            return apps.get_model(self.kwargs.get('app'), self.kwargs.get('model'))
        elif hasattr(self, 'get_queryset'):
            return self.get_queryset().model
        else:
            return None

    def get_model_config(self) -> Optional[ModelConfig]:
        """Get Trionyx model config"""
        if not hasattr(self, '__config'):
            setattr(self, '__config', models_config.get_config(self.get_model_class()))
        return getattr(self, '__config', None)


class ModelPermissionMixin:
    """Check Model permission"""

    permission_type = None

    def dispatch(self, request, *args, **kwargs):
        """Validate if user can use view"""
        if not self.is_enabled() or (self.permission and not request.user.has_perm(self.permission)):
            raise PermissionDenied()
        return super().dispatch(request, *args, **kwargs)

    def is_enabled(self):
        """Check if permission type is enabled for model"""
        if not hasattr(self, 'get_model_config'):
            return True

        if self.permission_type == 'change' and self.get_model_config().disable_change:
            return False

        if self.permission_type == 'add' and self.get_model_config().disable_add:
            return False

        if self.permission_type == 'delete' and self.get_model_config().disable_delete:
            return False

        return True

    @property
    def permission(self):
        """Permission for view"""
        if self.permission_type and hasattr(self, 'get_model_config'):
            config = self.get_model_config()

            return '{app_label}.{type}_{model_name}'.format(
                app_label=config.app_label,
                type=self.permission_type,
                model_name=config.model_name,
            ).lower()
        return None


class SessionValueMixin:
    """Mixin for handling session values"""

    def get_and_save_value(self, name, default=None):
        """Get value from request/session and save value to session"""
        if getattr(self, name, None):
            return getattr(self, name)

        value = self.get_session_value(name, default)
        value = self.request.POST.get(name, value)
        self.save_value(name, value)
        return value

    def get_session_value(self, name, default=None):
        """Get value from session"""
        session_name = 'list_{}_{}_{}'.format(self.kwargs.get('app'), self.kwargs.get('model'), name)
        return self.request.session.get(session_name, default)

    def save_value(self, name, value):
        """Save value to session"""
        session_name = 'list_{}_{}_{}'.format(self.kwargs.get('app'), self.kwargs.get('model'), name)
        self.request.session[session_name] = value
        setattr(self, name, value)
        return value



import time
from celery import shared_task
from django.utils import timezone
from trionyx.tasks import BaseTask

@shared_task
def manual_task(some_input):
    print(some_input)

@shared_task
def cron_task():
    print('run cron task')

class PostTask(BaseTask):
    name = 'post_publish'
    task_description = 'Post publish'

    def run(self):
        if not self.get_object():
            raise Exception('No object')

        for x in range(100):
            time.sleep(0.25)

            self.set_progress(x)

            if x % 10 == 0:
                self.add_output('Process at {}%'.format(x))

        object = self.get_object()
        object.publish_date = timezone.now()
        object.save()

        return object.publish_date


"""
trionyx.trionyx.context_processors
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: 2018 by Maikel Martens
:license: GPLv3
"""
from django.conf import settings
from trionyx.menu import app_menu
from trionyx import utils
from trionyx.urls import model_url
from trionyx.models import get_class


def trionyx(request):
    """Add trionyx context data"""
    return {
        'DEBUG': settings.DEBUG,
        'TX_APP_NAME': settings.TX_APP_NAME,
        'TX_LOGO_NAME_START': settings.TX_LOGO_NAME_START,
        'TX_LOGO_NAME_END': settings.TX_LOGO_NAME_END,
        'TX_LOGO_NAME_SMALL_START': settings.TX_LOGO_NAME_SMALL_START,
        'TX_LOGO_NAME_SMALL_END': settings.TX_LOGO_NAME_SMALL_END,
        'TX_THEME_COLOR': settings.TX_THEME_COLOR,
        'tx_skin_css': 'css/skins/skin-{}.min.css'.format(settings.TX_THEME_COLOR),
        'tx_tasks_url': model_url(get_class('trionyx.Task'), 'list'),

        'trionyx_menu_items': app_menu.get_menu_items(request.user),
        'trionyx_menu_collapse': request.COOKIES.get('menu.state') == 'collapsed',

        'datetime_input_format': utils.datetime_format_to_momentjs(utils.get_datetime_input_format()),
        'date_input_format': utils.datetime_format_to_momentjs(utils.get_datetime_input_format(date_only=True)),
        'current_locale': utils.get_current_locale(),
    }



"""
trionyx.data.currencies
~~~~~~~~~~~~~~~~~~~~~~~

:copyright: 2019 by Maikel Martens
:license: GPLv3
"""
CURRENCIES = (
    ("AFN", "Afghani"),
    ("EUR", "Euro"),
    ("ALL", "Lek"),
    ("DZD", "Algerian Dinar"),
    ("USD", "US Dollar"),
    ("AOA", "Kwanza"),
    ("ARS", "Argentine Peso"),
    ("AMD", "Armenian Dram"),
    ("AWG", "Aruban Florin"),
    ("AUD", "Australian Dollar"),
    ("AZN", "Azerbaijan Manat"),
    ("BSD", "Bahamian Dollar"),
    ("BHD", "Bahraini Dinar"),
    ("BDT", "Taka"),
    ("BBD", "Barbados Dollar"),
    ("BYN", "Belarusian Ruble"),
    ("BZD", "Belize Dollar"),
    ("BMD", "Bermudian Dollar"),
    ("INR", "Indian Rupee"),
    ("BTN", "Ngultrum"),
    ("BOB", "Boliviano"),
    ("BOV", "Mvdol"),
    ("BAM", "Convertible Mark"),
    ("BWP", "Pula"),
    ("NOK", "Norwegian Krone"),
    ("BRL", "Brazilian Real"),
    ("BND", "Brunei Dollar"),
    ("BGN", "Bulgarian Lev"),
    ("BIF", "Burundi Franc"),
    ("CVE", "Cabo Verde Escudo"),
    ("KHR", "Riel"),
    ("CAD", "Canadian Dollar"),
    ("KYD", "Cayman Islands Dollar"),
    ("CLP", "Chilean Peso"),
    ("CLF", "Unidad de Fomento"),
    ("CNY", "Yuan Renminbi"),
    ("COP", "Colombian Peso"),
    ("COU", "Unidad de Valor Real"),
    ("KMF", "Comorian Franc "),
    ("CDF", "Congolese Franc"),
    ("NZD", "New Zealand Dollar"),
    ("CRC", "Costa Rican Colon"),
    ("HRK", "Kuna"),
    ("CUP", "Cuban Peso"),
    ("CUC", "Peso Convertible"),
    ("ANG", "Netherlands Antillean Guilder"),
    ("CZK", "Czech Koruna"),
    ("DKK", "Danish Krone"),
    ("DJF", "Djibouti Franc"),
    ("DOP", "Dominican Peso"),
    ("EGP", "Egyptian Pound"),
    ("SVC", "El Salvador Colon"),
    ("ERN", "Nakfa"),
    ("ETB", "Ethiopian Birr"),
    ("FKP", "Falkland Islands Pound"),
    ("FJD", "Fiji Dollar"),
    ("GMD", "Dalasi"),
    ("GEL", "Lari"),
    ("GHS", "Ghana Cedi"),
    ("GIP", "Gibraltar Pound"),
    ("GTQ", "Quetzal"),
    ("GBP", "Pound Sterling"),
    ("GNF", "Guinean Franc"),
    ("GYD", "Guyana Dollar"),
    ("HTG", "Gourde"),
    ("HNL", "Lempira"),
    ("HKD", "Hong Kong Dollar"),
    ("HUF", "Forint"),
    ("ISK", "Iceland Krona"),
    ("IDR", "Rupiah"),
    ("IRR", "Iranian Rial"),
    ("IQD", "Iraqi Dinar"),
    ("ILS", "New Israeli Sheqel"),
    ("JMD", "Jamaican Dollar"),
    ("JPY", "Yen"),
    ("JOD", "Jordanian Dinar"),
    ("KZT", "Tenge"),
    ("KES", "Kenyan Shilling"),
    ("KPW", "North Korean Won"),
    ("KRW", "Won"),
    ("KWD", "Kuwaiti Dinar"),
    ("KGS", "Som"),
    ("LAK", "Lao Kip"),
    ("LBP", "Lebanese Pound"),
    ("LSL", "Loti"),
    ("ZAR", "Rand"),
    ("LRD", "Liberian Dollar"),
    ("LYD", "Libyan Dinar"),
    ("CHF", "Swiss Franc"),
    ("MOP", "Pataca"),
    ("MKD", "Denar"),
    ("MGA", "Malagasy Ariary"),
    ("MWK", "Malawi Kwacha"),
    ("MYR", "Malaysian Ringgit"),
    ("MVR", "Rufiyaa"),
    ("MRU", "Ouguiya"),
    ("MUR", "Mauritius Rupee"),
    ("MXN", "Mexican Peso"),
    ("MXV", "Mexican Unidad de Inversion (UDI)"),
    ("MDL", "Moldovan Leu"),
    ("MNT", "Tugrik"),
    ("MAD", "Moroccan Dirham"),
    ("MZN", "Mozambique Metical"),
    ("MMK", "Kyat"),
    ("NAD", "Namibia Dollar"),
    ("NPR", "Nepalese Rupee"),
    ("NIO", "Cordoba Oro"),
    ("NGN", "Naira"),
    ("OMR", "Rial Omani"),
    ("PKR", "Pakistan Rupee"),
    ("PAB", "Balboa"),
    ("PGK", "Kina"),
    ("PYG", "Guarani"),
    ("PEN", "Sol"),
    ("PHP", "Philippine Peso"),
    ("PLN", "Zloty"),
    ("QAR", "Qatari Rial"),
    ("RON", "Romanian Leu"),
    ("RUB", "Russian Ruble"),
    ("RWF", "Rwanda Franc"),
    ("SHP", "Saint Helena Pound"),
    ("WST", "Tala"),
    ("STN", "Dobra"),
    ("SAR", "Saudi Riyal"),
    ("RSD", "Serbian Dinar"),
    ("SCR", "Seychelles Rupee"),
    ("SLL", "Leone"),
    ("SGD", "Singapore Dollar"),
    ("SBD", "Solomon Islands Dollar"),
    ("SOS", "Somali Shilling"),
    ("SSP", "South Sudanese Pound"),
    ("LKR", "Sri Lanka Rupee"),
    ("SDG", "Sudanese Pound"),
    ("SRD", "Surinam Dollar"),
    ("SZL", "Lilangeni"),
    ("SEK", "Swedish Krona"),
    ("CHE", "WIR Euro"),
    ("CHW", "WIR Franc"),
    ("SYP", "Syrian Pound"),
    ("TWD", "New Taiwan Dollar"),
    ("TJS", "Somoni"),
    ("TZS", "Tanzanian Shilling"),
    ("THB", "Baht"),
    ("TOP", "Pa’anga"),
    ("TTD", "Trinidad and Tobago Dollar"),
    ("TND", "Tunisian Dinar"),
    ("TRY", "Turkish Lira"),
    ("TMT", "Turkmenistan New Manat"),
    ("UGX", "Uganda Shilling"),
    ("UAH", "Hryvnia"),
    ("AED", "UAE Dirham"),
    ("USN", "US Dollar (Next day)"),
    ("UYU", "Peso Uruguayo"),
    ("UYI", "Uruguay Peso en Unidades Indexadas (UI)"),
    ("UZS", "Uzbekistan Sum"),
    ("VUV", "Vatu"),
    ("VEF", "Bolívar"),
    ("VND", "Dong"),
    ("YER", "Yemeni Rial"),
    ("ZMW", "Zambian Kwacha"),
    ("ZWL", "Zimbabwe Dollar"),
)



from rest_framework import serializers
from trionyx.trionyx.models import User
from trionyx.api.serializers import register


@register(User)
class UserSerializer(serializers.ModelSerializer):
    class Meta:
        model = User
        fields = ['id', 'verbose_name']


"""
trionyx.utils
~~~~~~~~~~~~~~

:copyright: 2017 by Maikel Martens
:license: GPLv3
"""
import time
import logging
import random
import string
import importlib
import hashlib
from functools import reduce

from django.conf import settings
from django.utils import translation
from django.utils import formats
from django.core.cache import cache

logger = logging.getLogger(__name__)


class CacheLock:
    """CacheLock uses the django cache to give a lock for given keys"""

    def __init__(self, *keys, timeout=None):
        """Init CacheLock"""
        self.keys = keys
        self.timeout = timeout

    @property
    def cache_key(self):
        """Cache key"""
        return 'trionyx-cache-lock-{key}'.format(
            key=hashlib.md5(''.join([str(k) for k in self.keys]).encode()).hexdigest()
        )

    def __enter__(self):
        """Acquire lock"""
        run_time = 0
        wait_time = 0.1
        while not cache.add(self.cache_key, 'true', 999999):
            if self.timeout and run_time > self.timeout:
                raise TimeoutError()
            time.sleep(wait_time)
            run_time += wait_time

    def __exit__(self, type, value, traceback):
        """Clear cache lock"""
        cache.delete(self.cache_key)


def random_string(size):
    """Create random string containing ascii leters and digits, with the length of given size"""
    return ''.join(random.choice(string.ascii_letters + string.digits) for _ in range(size))


def import_object_by_string(namespace):
    """Import object by complete namespace"""
    segments = namespace.split('.')
    module = importlib.import_module('.'.join(segments[:-1]))
    return getattr(module, segments[-1])


def create_celerybeat_schedule(apps):
    """Create Celery beat schedule by get schedule from every installed app"""
    beat_schedule = {}
    for app in apps:
        try:
            config = import_object_by_string(app)
            module = importlib.import_module('{}.cron'.format(config.name))
        except Exception:
            try:
                module = importlib.import_module('{}.cron'.format(app))
            except Exception:
                continue

        if not (hasattr(module, 'schedule') and isinstance(module.schedule, dict)):
            logger.warning('{} has no schedule or schedule is not a dict'.format(module.__name__))
            continue

        # Add cron queue option
        for name, schedule in module.schedule.items():
            options = schedule.get('options', {})
            if 'queue' not in options:
                options['queue'] = 'cron'
                schedule['options'] = options

                beat_schedule[name] = schedule

    return beat_schedule


def get_current_language():
    """Get active language by django.utils.translation or return settings LANGUAGE_CODE"""
    if translation.get_language():
        return translation.get_language()
    return settings.LANGUAGE_CODE


def get_current_locale():
    """Get active locale based on get_current_language function"""
    return translation.to_locale(get_current_language())


def get_datetime_input_format(date_only=False):
    """
    Get current locale date input format.

    :param date_only:
    :return:
    """
    return formats.get_format_lazy('DATE_INPUT_FORMATS' if date_only else 'DATETIME_INPUT_FORMATS')[0]


def datetime_format_to_momentjs(format):
    """
    Convert python datetime string format to momentjs string format.

    :param format: python datetime string format
    :return:
    """
    mapping = {
        '%a': 'ddd',
        '%A': 'dddd',
        '%w': 'd',
        '%d': 'DD',
        '%b': 'MMM',
        '%B': 'MMMM',
        '%m': 'MM',
        '%y': 'YY',
        '%Y': 'YYYY',
        '%H': 'HH',
        '%I': 'hh',
        '%p': 'A',
        '%M': 'mm',
        '%S': 'ss',
        '%f': 'SSS',
        '%z': 'ZZ',
        '%Z': 'z',
        '%j': 'DDDD',
        '%U': 'ww',
        '%W': 'ww',
        '%c': 'ddd MMM DD HH:mm:ss YYYY',
        '%x': 'MM/DD/YYYY',
        '%X': 'HH:mm:ss',
        '%%': '%'
    }
    return reduce(lambda value, key: value.replace(key, mapping[key]), mapping.keys(), format)


def datetime_format_to_django_template(format):
    """
    Convert python datetime string format to django template string format.

    :param format: python datetime string format
    :return:
    """
    mapping = {
        '%d': 'd',  # Day of the month, 2 digits with leading zeros.
        '%m': 'm',  # Month, 2 digits with leading zeros.
        '%y': 'y',  # Year, 2 digits.
        '%Y': 'Y',  # Year, 4 digits.
        '%H': 'H',  # Hour, 24-hour format.
        '%M': 'i',  # Minutes
        '%S': 's',  # Seconds, 2 digits with leading zeros.
        '%p': 'A',  # 'AM' or 'PM'
    }
    return reduce(lambda value, key: value.replace(key, mapping[key]), mapping.keys(), format)


def get_current_request():
    """Get current request object"""
    from trionyx.trionyx.middleware import LOCAL_DATA
    return getattr(LOCAL_DATA, 'request', None)



# Copyright 2019 by Maikel Martens
#
# License GPLv3

"""
Settings
========

All Trionyx base settings
"""
import json
import logging
import os
from pkg_resources import iter_entry_points

from django.core.exceptions import ImproperlyConfigured
from kombu import Queue, Exchange


def gettext_noop(s):
    """Return same string, Dummy function to find translatable strings with makemessages"""
    return s


try:
    with open(os.path.abspath(os.environ.get('TRIONYX_CONFIG', 'environment.json'))) as f:
        trionyx_config = json.loads(f.read())
except FileNotFoundError:
    raise ImproperlyConfigured(
        "Could not load Trionyx config file, is env variable TRIONYX_CONFIG correctly configuerd?")


def get_env_var(setting, default=None, configs=trionyx_config):
    """
    Get environment variable from the environment json file

    Default environment file is `environment.json` in the root of project,
    Other file path can be set with the `TRIONYX_CONFIG` environment variable
    """
    try:
        return configs[setting]
    except KeyError:
        if default is not None:
            return default
        raise ImproperlyConfigured("ImproperlyConfigured: Set {} environment variable".format(setting))


DEBUG = get_env_var('DEBUG', False)
ALLOWED_HOSTS = get_env_var('ALLOWED_HOSTS', [])
SECRET_KEY = get_env_var('SECRET_KEY')

INSTALLED_APPS = [
    # Trionyx core app
    'trionyx.trionyx',

    # Add installed trionyx apps
    *[entry_point.module_name for entry_point in iter_entry_points(group='trionyx.app', name=None)],

    # Django apps
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'django.contrib.humanize',

    'rest_framework',
    'rest_framework.authtoken',
    'watson',
    'crispy_forms',
    'compressor',
]

# Internationalization
# https://docs.djangoproject.com/en/1.11/topics/i18n/

LANGUAGE_CODE = 'en-us'
TIME_ZONE = 'Etc/GMT'
USE_I18N = True
USE_L10N = True
USE_TZ = True

LANGUAGES = [
    ('en-us', gettext_noop('English')),
    ('nl', gettext_noop('Dutch')),
]

# ==============================================================================
# Project URLS and media settings
# ==============================================================================
ROOT_URLCONF = 'trionyx.urls'

LOGIN_URL = 'trionyx:login'
LOGOUT_URL = 'trionyx:logout'
LOGIN_REDIRECT_URL = '/'

STATIC_URL = '/static/'
MEDIA_URL = '/media/'

WSGI_APPLICATION = 'wsgi.application'

# ==============================================================================
# Middleware
# ==============================================================================
MIDDLEWARE = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'trionyx.trionyx.middleware.LocalizationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'trionyx.trionyx.middleware.LoginRequiredMiddleware',
    'trionyx.trionyx.middleware.GlobalRequestMiddleware',
    'trionyx.trionyx.middleware.LastLoginMiddleware',
)

# ==============================================================================
# Auth / security
# ==============================================================================
AUTH_USER_MODEL = 'trionyx.User'

SECURE_PROXY_SSL_HEADER = ('HTTP_X_FORWARDED_PROTO', 'https')

LOGIN_EXEMPT_URLS = [
    'static',
    'api',
]
"""A list of urls that dont require a login"""

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]

# ==============================================================================
# Templates
# ==============================================================================

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.contrib.auth.context_processors.auth',
                'django.template.context_processors.debug',
                'django.template.context_processors.i18n',
                'django.template.context_processors.media',
                'django.template.context_processors.static',
                'django.template.context_processors.tz',
                'django.template.context_processors.request',
                'django.contrib.messages.context_processors.messages',

                'django.contrib.messages.context_processors.messages',
                'trionyx.trionyx.context_processors.trionyx',
            ],
        }
    },
]

CRISPY_TEMPLATE_PACK = 'bootstrap3'

# ==============================================================================
# Compressor
# ==============================================================================
STATICFILES_FINDERS = [
    'django.contrib.staticfiles.finders.FileSystemFinder',
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
    'compressor.finders.CompressorFinder',
]

COMPRESS_ENABLED = True
COMPRESS_OFFLINE = True

# ==============================================================================
# Cache backend
# ==============================================================================
CACHES = {
    'default': {
        'BACKEND': 'django.core.cache.backends.db.DatabaseCache',
        'LOCATION': 'django_cache',
    }
}

CORS_ORIGIN_WHITELIST = [
    'localhost:8000',
]

# ==============================================================================
# Celery
# ==============================================================================
CELERY_BROKER_URL = 'amqp://guest:guest@localhost:5672//'

CELERY_ACCEPT_CONTENT = ['pickle']
CELERY_TASK_SERIALIZER = 'pickle'
CELERY_RESULT_SERIALIZER = 'pickle'
CELERY_ENABLE_UTC = True

CELERY_QUEUES = (
    Queue('cron', Exchange('cron'), routing_key='cron'),
    Queue('low_prio', Exchange('low_prio'), routing_key='low_prio'),
    Queue('high_prio', Exchange('high_prio'), routing_key='high_prio'),
)
CELERY_TASK_DEFAULT_QUEUE = 'low_prio'

CELERY_TASK_SOFT_TIME_LIMIT = 3600
CELERY_TASK_TIME_LIMIT = 3900

# ==============================================================================
# Trionyx settings
# ==============================================================================
REST_FRAMEWORK = {
    'DEFAULT_FILTER_BACKENDS': (
        'trionyx.api.filters.SearchFilter',
        'rest_framework.filters.OrderingFilter',
    ),
    'DEFAULT_RENDERER_CLASSES': (
        'rest_framework.renderers.BrowsableAPIRenderer',
        'rest_framework.renderers.JSONRenderer',
    ),
    'DEFAULT_PERMISSION_CLASSES': (
        'rest_framework.permissions.IsAuthenticated',
    ),
    'DEFAULT_AUTHENTICATION_CLASSES': (
        'rest_framework.authentication.TokenAuthentication',
        'rest_framework.authentication.BasicAuthentication',
        'rest_framework.authentication.SessionAuthentication',
    ),
    'DEFAULT_PAGINATION_CLASS': 'trionyx.api.pagination.PageNumberPagination',
    'PAGE_SIZE': 20,
}

# ==============================================================================
# Trionyx settings
# ==============================================================================
TX_APP_NAME = 'Trionyx'
"""Full application name"""

TX_LOGO_NAME_START = 'Tri'
"""The first characters of the name that are bold"""

TX_LOGO_NAME_END = 'onyx'
"""The rest of the characters"""

TX_LOGO_NAME_SMALL_START = 'T'
"""The first character or characters of the small logo that is bold"""

TX_LOGO_NAME_SMALL_END = 'X'
"""The last character or characters of the small logo that is normal"""

TX_THEME_COLOR = 'purple'
"""The theme skin color (header). Aviable colors: blue, yellow, green, purple, red, black. All colors have a light version blue-light"""


def TX_DEFAULT_DASHBOARD():
    """Return default dashboard"""
    from django.contrib.contenttypes.models import ContentType
    from django.utils.translation import ugettext as _
    return [
        {
            "code": "auditlog",
            "config": {
                "title": _("Action history"),
                "color": "light-blue",
                "show": "all",
                "refresh": "0"
            },
            "x": 0,
            "y": 5,
            "w": 6,
            "h": 22,
        },
        {
            "code": "total_summary",
            "config": {
                "title": _("Unique users today"),
                "color": "purple",
                "model": ContentType.objects.get_by_natural_key('trionyx', 'user').id,
                "field": "__count__",
                "icon": "fa fa-user",
                "period": "day",
                "period_field": "last_online",
                "refresh": "15"
            },
            "x": 0,
            "y": 0,
            "w": 4,
            "h": 5,
        },
        {
            "code": "total_summary",
            "config": {
                "title": _("New users this week"),
                "color": "green",
                "refresh": "15",
                "icon": "fa fa-user-plus",
                "model": ContentType.objects.get_by_natural_key('trionyx', 'user').id,
                "field": "__count__",
                "period": "week",
                "period_field": "created_at"
            },
            "x": 4,
            "y": 0,
            "w": 4,
            "h": 5,
        },
        {
            "code": "total_summary",
            "config": {
                "title": _("User count"),
                "color": "yellow",
                "refresh": "0",
                "icon": "fa fa-users",
                "model": ContentType.objects.get_by_natural_key('trionyx', 'user').id,
                "field": "__count__",
                "period": "all",
                "period_field": ""
            },
            "x": 8,
            "y": 0,
            "w": 4,
            "h": 5,
        }
    ]


TX_MODEL_OVERWRITES = {}
"""
Config to overwrite models, its a dict where the key is the original `app_label.model_name` and value is the new one.

.. code-block:: python

    TX_MODEL_OVERWRITES = {
        'trionyx.User': 'local.User',
    }
"""

TX_MODEL_CONFIGS = {}
"""
Dict with configs for non Trionyx model, example:

.. code-block:: python

    TX_MODEL_CONFIGS = {
        'auth.group': {
            'list_default_fields': ['name'],
            'disable_search_index': False,
        }
    }
"""

TX_CORE_MODEL_CONFIGS = {
    'auth.group': {
        'list_default_fields': ['name'],
        'disable_search_index': False,
    },
    'watson.searchentry': {
        'hide_permissions': True,
    },
    'trionyx.userattribute': {
        'hide_permissions': True,
    },
    'trionyx.logentry': {
        'hide_permissions': True,
    },
    'trionyx.auditlogentry': {
        'hide_permissions': True,
    },
    'sessions.session': {
        'hide_permissions': True,
    },
    'contenttypes.contenttype': {
        'hide_permissions': True,
    },
    'authtoken.token': {
        'hide_permissions': True,
    },
    'auth.permission': {
        'hide_permissions': True,
    },
}

TX_DB_LOG_LEVEL = logging.WARNING
"""The DB log level for logging"""



"""
trionyx.models
~~~~~~~~~~~~~~

:copyright: 2018 by Maikel Martens
:license: GPLv3
"""
import operator
from collections import defaultdict
from functools import reduce

from django.conf import settings
from django.db.models import *  # noqa F403
from django.urls import reverse
from jsonfield import JSONField  # noqa F401
from django.contrib import messages
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _

from trionyx.config import models_config
from trionyx import utils

TX_MODEL_OVERWRITES = {key.lower(): value.lower() for key, value in settings.TX_MODEL_OVERWRITES.items()}


def get_name(model):
    """Get model name"""
    name = models_config.get_model_name(model)
    return TX_MODEL_OVERWRITES.get(name, name)


def get_class(model):
    """Get model class"""
    return models_config.get_config(get_name(model)).model


# =============================================================================
# Fields
# =============================================================================
class PriceField(DecimalField):  # noqa F405
    """Price field is Decimal(11,4) field, has no extra logic only for rendering"""

    def __init__(self, *args, **kwargs):
        """Init PriceField"""
        kwargs['max_digits'] = 11
        kwargs['decimal_places'] = 4
        super().__init__(*args, **kwargs)


# =============================================================================
# Base models
# =============================================================================
class BaseManager(Manager):  # noqa F405
    """model base manager for all Trionyx models"""

    def get_queryset(self):
        """Give qeuryset where deleted items are filtered"""
        return super().get_queryset().filter(deleted=False)


class BaseModel(Model):  # noqa F405
    """Base model for all Trionyx models"""

    objects = BaseManager()

    created_at = DateTimeField(_('Created at'), auto_now_add=True)  # noqa F405
    """Created at field, date is set when model is created"""

    updated_at = DateTimeField(_('Updated at'), auto_now=True)  # noqa F405
    """Update at field, date is set when model is saved"""

    deleted = BooleanField(_('Deleted'), default=False)  # noqa F405
    """Deleted field, object is soft deleted"""

    created_by = ForeignKey(
        get_name('trionyx.User'), SET_NULL, default=None, blank=True, null=True,
        related_name='+', verbose_name=_('Created by'))
    """Created by field"""

    verbose_name = TextField(_('Verbose name'), default='', blank=True)
    """Verbose name field"""

    class Meta:
        """Meta information for BaseModel"""

        abstract = True

    def __str__(self):
        """Give verbose name of object"""
        return self.verbose_name if self.verbose_name else self.generate_verbose_name()

    def save(self, *args, **kwargs):
        """Save model"""
        try:
            self.verbose_name = self.generate_verbose_name()
        except Exception:
            pass

        try:
            if not self.pk and not self.created_by:
                self.created_by = utils.get_current_request().user
        except Exception:
            pass

        return super().save(*args, **kwargs)

    def generate_verbose_name(self):
        """Generate verbose name"""
        from trionyx.renderer import LazyFieldRenderer
        app_label = self._meta.app_label
        model_name = type(self).__name__
        verbose_name = models_config.get_config(self).verbose_name
        return verbose_name.format(model_name=model_name, app_label=app_label, **{
            field.name: LazyFieldRenderer(self, field.name)
            for field in models_config.get_config(self).get_fields(True, True)
        })

    def get_absolute_url(self):
        """Get model url"""
        return reverse('trionyx:model-view', kwargs={
            'app': self._meta.app_label,
            'model': self._meta.model_name,
            'pk': self.id
        })


def filter_queryset_with_user_filters(queryset, filters, request=None):
    """Apply user provided filters on queryset"""
    config = models_config.get_config(queryset.model)

    field_indexed = {
        name: {
            'name': name,
            'label': field['label'],
            'type': field['type'],
            'choices': field['choices'],
        }
        for name, field in config.get_list_fields().items()
    }
    grouped_filter = defaultdict(list)
    for filter in filters:
        field = field_indexed.get(filter['field'])

        if not field:
            continue

        try:
            if filter['operator'] == 'null':
                queryset = queryset.filter(**{'{}__isnull'.format(filter['field']): filter['value']})
            elif field['type'] == 'datetime':
                filter['value'] = timezone.make_aware(
                    timezone.datetime.strptime(filter['value'], utils.get_datetime_input_format()))
            elif field['type'] == 'date':
                filter['value'] = timezone.make_aware(timezone.datetime.strptime(
                    filter['value'],
                    utils.get_datetime_input_format(date_only=True)))

            if filter['operator'] == '==':
                grouped_filter[filter['field']].append(filter['value'])
            elif filter['operator'] == '!=':
                if field['type'] == 'text':
                    queryset = queryset.exclude(**{'{}__icontains'.format(filter['field']): filter['value']})
                else:
                    queryset = queryset.exclude(**{filter['field']: filter['value']})
            elif filter['operator'] == '<':
                queryset = queryset.filter(**{'{}__lt'.format(filter['field']): filter['value']})
            elif filter['operator'] == '<=':
                queryset = queryset.filter(**{'{}__lte'.format(filter['field']): filter['value']})
            elif filter['operator'] == '>':
                queryset = queryset.filter(**{'{}__gt'.format(filter['field']): filter['value']})
            elif filter['operator'] == '>':
                queryset = queryset.filter(**{'{}__gte'.format(filter['field']): filter['value']})
        except Exception:
            if request:
                messages.add_message(request, messages.ERROR, "Could not apply filter ({} {} {})".format(
                    filter['field'],
                    filter['operator'],
                    filter['value']
                ))
    for name, values in grouped_filter.items():
        field = field_indexed[name]
        if field['type'] == 'text':
            or_queries = [Q(**{'{}__icontains'.format(name): value}) for value in values]
            queryset = queryset.filter(reduce(operator.or_, or_queries))
        else:
            or_queries = [Q(**{name: value}) for value in values]
            queryset = queryset.filter(reduce(operator.or_, or_queries))

    return queryset



"""
trionyx.tasks
~~~~~~~~~~~~~

:copyright: 2018 by Maikel Martens
:license: GPLv3
"""
import logging
from datetime import datetime

import celery
from celery import shared_task, current_app  # noqa F401
from celery.utils import uuid
from django.utils import timezone
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.utils.translation import ugettext_lazy as _

from trionyx.models import get_class
from trionyx.utils import CacheLock, get_current_request

Task = get_class('trionyx.Task')

logger = logging.getLogger(__name__)


class TaskMetaClass(type):
    """MetaClass for task"""

    def __new__(cls, clsname, bases, attrs):
        """Auto register celery task"""
        TaskClass = super().__new__(cls, clsname, bases, attrs)

        if clsname != 'BaseTask':
            current_app.tasks.register(TaskClass())

        return TaskClass


class BaseTask(celery.Task, metaclass=TaskMetaClass):
    """Trionyx BaseTask"""

    name = None
    task_queue = None
    task_description = _('Base Task')
    task_lock = True
    task_model = None

    def __call__(self, *args, **kwargs):
        """Run task"""
        self.__task, _ = Task.objects.get_or_create(celery_task_id=self.request.id)

        if self.__task.object_id and self.task_lock:
            with CacheLock('TASK_LOCK', self.__task.object_type_id, self.__task.object_id, timeout=settings.CELERY_TASK_TIME_LIMIT + 60):
                return self._run(*args, **kwargs)
        else:
            return self._run(*args, **kwargs)

    def _run(self, *args, **kwargs):
        """Run task and save result"""
        self.__task.status = Task.RUNNING
        self.__task.started_at = timezone.now()
        self.__task.save()

        try:
            if self.__task.user:
                with self.__task.user.locale_override():
                    result = super().__call__(*args, **kwargs)
                    if not result:
                        result = str(_('Task completed'))
            else:
                result = super().__call__(*args, **kwargs)
                if not result:
                    result = _('Task completed')

            duration = timezone.now() - self.__task.started_at
            self.__task.execution_time = int(duration.total_seconds())
            self.__task.result = str(result)
            self.__task.progress = 100
            self.__task.status = Task.COMPLETE
            self.__task.save()
        except Exception as e:
            logger.exception(e)
            result = str(e)

            self.__task.result = result
            self.__task.status = Task.FAILED
            self.__task.save()

        return result

    def delay(self, *args, **kwargs):
        """Start task"""
        task_id = uuid()
        user = kwargs.pop('task_user', None)
        description = kwargs.pop('task_description', self.task_description)
        object = kwargs.pop('task_object', None)
        model = kwargs.pop('task_model', self.task_model)
        eta = kwargs.pop('task_eta', None)
        queue = kwargs.pop('task_queue', self.task_queue)

        eta = eta if isinstance(eta, datetime) else None
        model = object if object else model

        if not user and get_current_request() and get_current_request().user.is_authenticated:
            user = get_current_request().user

        Task.objects.create(
            celery_task_id=task_id,
            description=description,
            identifier=self.name,
            object_id=object.id if object else None,
            object_type=ContentType.objects.get_for_model(model) if model else None,
            object_verbose_name=str(object) if object else '',
            user=user,
            status=Task.SCHEDULED if eta else Task.QUEUE,
            scheduled_at=eta
        )

        return self.apply_async(args=args, kwargs=kwargs, task_id=task_id, eta=eta, queue=queue)

    def set_progress(self, progress):
        """Set progress"""
        progress = progress if progress > 0 else 0
        progress = progress if progress <= 100 else 100

        self.__task.progress = progress
        self.__task.save()

    def add_output(self, output):
        """Add task process output"""
        logger.info('TASK OUTPUT: {output}'.format(output=output))
        self.__task.progress_output.append(output)
        self.__task.save()

    def get_user(self):
        """Return task user"""
        return self.__task.user

    def get_object(self):
        """Return task object"""
        return self.__task.object

    def get_model(self):
        """Return task model"""
        if self.__task.object_type:
            return self.__task.object_type.model_class()
        return None



from trionyx import models

class Category(models.BaseModel):
    name = models.CharField(max_length=255)
    description = models.TextField(default='')

    class Meta:
        verbose_name_plural = 'Categories'


class Tag(models.BaseModel):
    name = models.CharField(max_length=255)


class Post(models.BaseModel):
    title = models.CharField(max_length=255)
    content = models.TextField()

    publish_date = models.DateTimeField()

    category = models.ForeignKey(Category, models.CASCADE, related_name='posts')
    tags = models.ManyToManyField(Tag, related_name='posts')

    STATUS_NEW = 1
    STATUS_DRAFT = 2
    STATUS_PUBLISHED = 3

    STATUS_CHOISES = [
        (STATUS_NEW, 'New'),
        (STATUS_DRAFT, 'Draft'),
        (STATUS_PUBLISHED, 'Published'),
    ]

    status = models.IntegerField(choices=STATUS_CHOISES, default=STATUS_NEW)

    price = models.PriceField(default=0.0)



# Generated by Django 2.2.6 on 2019-10-29 19:50

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import jsonfield.encoder
import jsonfield.fields
import trionyx.trionyx.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('auth', '0011_update_proxy_permissions'),
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='User',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('password', models.CharField(max_length=128, verbose_name='password')),
                ('last_login', models.DateTimeField(blank=True, null=True, verbose_name='last login')),
                ('is_superuser', models.BooleanField(default=False, help_text='Designates that this user has all permissions without explicitly assigning them.', verbose_name='superuser status')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='Created at')),
                ('updated_at', models.DateTimeField(auto_now=True, verbose_name='Updated at')),
                ('deleted', models.BooleanField(default=False, verbose_name='Deleted')),
                ('verbose_name', models.TextField(blank=True, default='', verbose_name='Verbose name')),
                ('email', models.EmailField(max_length=255, unique=True, verbose_name='Email')),
                ('first_name', models.CharField(blank=True, default='', max_length=64, verbose_name='First name')),
                ('last_name', models.CharField(blank=True, default='', max_length=64, verbose_name='Last name')),
                ('is_active', models.BooleanField(default=True, verbose_name='Active')),
                ('date_joined', models.DateTimeField(default=django.utils.timezone.now, verbose_name='Date joined')),
                ('last_online', models.DateTimeField(blank=None, null=True, verbose_name='Last online')),
                ('avatar', models.ImageField(blank=True, default='', upload_to='avatars/', verbose_name='Avatar')),
                ('language', models.CharField(choices=[('en-us', 'English'), ('nl', 'Dutch')], default=trionyx.trionyx.models.default_language, max_length=6, verbose_name='Language')),
                ('timezone', models.CharField(choices=[('Etc/GMT+12', 'Etc/GMT+12'), ('Etc/GMT+11', 'Etc/GMT+11'), ('Pacific/Midway', 'Pacific/Midway'), ('Pacific/Niue', 'Pacific/Niue'), ('Pacific/Pago_Pago', 'Pacific/Pago_Pago'), ('Etc/GMT+10', 'Etc/GMT+10'), ('Pacific/Honolulu', 'Pacific/Honolulu'), ('Pacific/Johnston', 'Pacific/Johnston'), ('Pacific/Rarotonga', 'Pacific/Rarotonga'), ('Pacific/Tahiti', 'Pacific/Tahiti'), ('America/Anchorage', 'America/Anchorage'), ('America/Juneau', 'America/Juneau'), ('America/Nome', 'America/Nome'), ('America/Sitka', 'America/Sitka'), ('America/Yakutat', 'America/Yakutat'), ('America/Santa_Isabel', 'America/Santa_Isabel'), ('America/Dawson', 'America/Dawson'), ('America/Los_Angeles', 'America/Los_Angeles'), ('America/Tijuana', 'America/Tijuana'), ('America/Vancouver', 'America/Vancouver'), ('America/Whitehorse', 'America/Whitehorse'), ('PST8PDT', 'PST8PDT'), ('America/Creston', 'America/Creston'), ('America/Dawson_Creek', 'America/Dawson_Creek'), ('America/Hermosillo', 'America/Hermosillo'), ('America/Phoenix', 'America/Phoenix'), ('Etc/GMT+7', 'Etc/GMT+7'), ('America/Chihuahua', 'America/Chihuahua'), ('America/Mazatlan', 'America/Mazatlan'), ('America/Boise', 'America/Boise'), ('America/Cambridge_Bay', 'America/Cambridge_Bay'), ('America/Denver', 'America/Denver'), ('America/Edmonton', 'America/Edmonton'), ('America/Inuvik', 'America/Inuvik'), ('America/Ojinaga', 'America/Ojinaga'), ('America/Yellowknife', 'America/Yellowknife'), ('MST7MDT', 'MST7MDT'), ('America/Belize', 'America/Belize'), ('America/Costa_Rica', 'America/Costa_Rica'), ('America/El_Salvador', 'America/El_Salvador'), ('America/Guatemala', 'America/Guatemala'), ('America/Managua', 'America/Managua'), ('America/Tegucigalpa', 'America/Tegucigalpa'), ('Etc/GMT+6', 'Etc/GMT+6'), ('Pacific/Galapagos', 'Pacific/Galapagos'), ('America/Chicago', 'America/Chicago'), ('America/Indiana/Knox', 'America/Indiana/Knox'), ('America/Indiana/Tell_City', 'America/Indiana/Tell_City'), ('America/Matamoros', 'America/Matamoros'), ('America/Menominee', 'America/Menominee'), ('America/North_Dakota/Beulah', 'America/North_Dakota/Beulah'), ('America/North_Dakota/Center', 'America/North_Dakota/Center'), ('America/North_Dakota/New_Salem', 'America/North_Dakota/New_Salem'), ('America/Rainy_River', 'America/Rainy_River'), ('America/Rankin_Inlet', 'America/Rankin_Inlet'), ('America/Resolute', 'America/Resolute'), ('America/Winnipeg', 'America/Winnipeg'), ('CST6CDT', 'CST6CDT'), ('America/Bahia_Banderas', 'America/Bahia_Banderas'), ('America/Cancun', 'America/Cancun'), ('America/Merida', 'America/Merida'), ('America/Mexico_City', 'America/Mexico_City'), ('America/Monterrey', 'America/Monterrey'), ('America/Regina', 'America/Regina'), ('America/Swift_Current', 'America/Swift_Current'), ('America/Bogota', 'America/Bogota'), ('America/Cayman', 'America/Cayman'), ('America/Coral_Harbour', 'America/Coral_Harbour'), ('America/Eirunepe', 'America/Eirunepe'), ('America/Guayaquil', 'America/Guayaquil'), ('America/Jamaica', 'America/Jamaica'), ('America/Lima', 'America/Lima'), ('America/Panama', 'America/Panama'), ('America/Rio_Branco', 'America/Rio_Branco'), ('Etc/GMT+5', 'Etc/GMT+5'), ('America/Detroit', 'America/Detroit'), ('America/Havana', 'America/Havana'), ('America/Indiana/Petersburg', 'America/Indiana/Petersburg'), ('America/Indiana/Vincennes', 'America/Indiana/Vincennes'), ('America/Indiana/Winamac', 'America/Indiana/Winamac'), ('America/Iqaluit', 'America/Iqaluit'), ('America/Kentucky/Monticello', 'America/Kentucky/Monticello'), ('America/Louisville', 'America/Louisville'), ('America/Montreal', 'America/Montreal'), ('America/Nassau', 'America/Nassau'), ('America/New_York', 'America/New_York'), ('America/Nipigon', 'America/Nipigon'), ('America/Pangnirtung', 'America/Pangnirtung'), ('America/Port-au-Prince', 'America/Port-au-Prince'), ('America/Thunder_Bay', 'America/Thunder_Bay'), ('America/Toronto', 'America/Toronto'), ('EST5EDT', 'EST5EDT'), ('America/Indiana/Marengo', 'America/Indiana/Marengo'), ('America/Indiana/Vevay', 'America/Indiana/Vevay'), ('America/Indianapolis', 'America/Indianapolis'), ('America/Caracas', 'America/Caracas'), ('America/Asuncion', 'America/Asuncion'), ('America/Glace_Bay', 'America/Glace_Bay'), ('America/Goose_Bay', 'America/Goose_Bay'), ('America/Halifax', 'America/Halifax'), ('America/Moncton', 'America/Moncton'), ('America/Thule', 'America/Thule'), ('Atlantic/Bermuda', 'Atlantic/Bermuda'), ('America/Campo_Grande', 'America/Campo_Grande'), ('America/Cuiaba', 'America/Cuiaba'), ('America/Anguilla', 'America/Anguilla'), ('America/Antigua', 'America/Antigua'), ('America/Aruba', 'America/Aruba'), ('America/Barbados', 'America/Barbados'), ('America/Blanc-Sablon', 'America/Blanc-Sablon'), ('America/Boa_Vista', 'America/Boa_Vista'), ('America/Curacao', 'America/Curacao'), ('America/Dominica', 'America/Dominica'), ('America/Grand_Turk', 'America/Grand_Turk'), ('America/Grenada', 'America/Grenada'), ('America/Guadeloupe', 'America/Guadeloupe'), ('America/Guyana', 'America/Guyana'), ('America/Kralendijk', 'America/Kralendijk'), ('America/La_Paz', 'America/La_Paz'), ('America/Lower_Princes', 'America/Lower_Princes'), ('America/Manaus', 'America/Manaus'), ('America/Marigot', 'America/Marigot'), ('America/Martinique', 'America/Martinique'), ('America/Montserrat', 'America/Montserrat'), ('America/Port_of_Spain', 'America/Port_of_Spain'), ('America/Porto_Velho', 'America/Porto_Velho'), ('America/Puerto_Rico', 'America/Puerto_Rico'), ('America/Santo_Domingo', 'America/Santo_Domingo'), ('America/St_Barthelemy', 'America/St_Barthelemy'), ('America/St_Kitts', 'America/St_Kitts'), ('America/St_Lucia', 'America/St_Lucia'), ('America/St_Thomas', 'America/St_Thomas'), ('America/St_Vincent', 'America/St_Vincent'), ('America/Tortola', 'America/Tortola'), ('Etc/GMT+4', 'Etc/GMT+4'), ('America/Santiago', 'America/Santiago'), ('Antarctica/Palmer', 'Antarctica/Palmer'), ('America/St_Johns', 'America/St_Johns'), ('America/Sao_Paulo', 'America/Sao_Paulo'), ('America/Argentina/La_Rioja', 'America/Argentina/La_Rioja'), ('America/Argentina/Rio_Gallegos', 'America/Argentina/Rio_Gallegos'), ('America/Argentina/Salta', 'America/Argentina/Salta'), ('America/Argentina/San_Juan', 'America/Argentina/San_Juan'), ('America/Argentina/San_Luis', 'America/Argentina/San_Luis'), ('America/Argentina/Tucuman', 'America/Argentina/Tucuman'), ('America/Argentina/Ushuaia', 'America/Argentina/Ushuaia'), ('America/Buenos_Aires', 'America/Buenos_Aires'), ('America/Catamarca', 'America/Catamarca'), ('America/Cordoba', 'America/Cordoba'), ('America/Jujuy', 'America/Jujuy'), ('America/Mendoza', 'America/Mendoza'), ('America/Araguaina', 'America/Araguaina'), ('America/Belem', 'America/Belem'), ('America/Cayenne', 'America/Cayenne'), ('America/Fortaleza', 'America/Fortaleza'), ('America/Maceio', 'America/Maceio'), ('America/Paramaribo', 'America/Paramaribo'), ('America/Recife', 'America/Recife'), ('America/Santarem', 'America/Santarem'), ('Antarctica/Rothera', 'Antarctica/Rothera'), ('Atlantic/Stanley', 'Atlantic/Stanley'), ('Etc/GMT+3', 'Etc/GMT+3'), ('America/Godthab', 'America/Godthab'), ('America/Montevideo', 'America/Montevideo'), ('America/Bahia', 'America/Bahia'), ('America/Noronha', 'America/Noronha'), ('Atlantic/South_Georgia', 'Atlantic/South_Georgia'), ('Etc/GMT+2', 'Etc/GMT+2'), ('America/Scoresbysund', 'America/Scoresbysund'), ('Atlantic/Azores', 'Atlantic/Azores'), ('Atlantic/Cape_Verde', 'Atlantic/Cape_Verde'), ('Etc/GMT+1', 'Etc/GMT+1'), ('Africa/Casablanca', 'Africa/Casablanca'), ('Africa/El_Aaiun', 'Africa/El_Aaiun'), ('America/Danmarkshavn', 'America/Danmarkshavn'), ('Etc/GMT', 'Etc/GMT'), ('Europe/Isle_of_Man', 'Europe/Isle_of_Man'), ('Europe/Guernsey', 'Europe/Guernsey'), ('Europe/Jersey', 'Europe/Jersey'), ('Europe/London', 'Europe/London'), ('Atlantic/Canary', 'Atlantic/Canary'), ('Atlantic/Faeroe', 'Atlantic/Faeroe'), ('Atlantic/Madeira', 'Atlantic/Madeira'), ('Europe/Dublin', 'Europe/Dublin'), ('Europe/Lisbon', 'Europe/Lisbon'), ('Africa/Abidjan', 'Africa/Abidjan'), ('Africa/Accra', 'Africa/Accra'), ('Africa/Bamako', 'Africa/Bamako'), ('Africa/Banjul', 'Africa/Banjul'), ('Africa/Bissau', 'Africa/Bissau'), ('Africa/Conakry', 'Africa/Conakry'), ('Africa/Dakar', 'Africa/Dakar'), ('Africa/Freetown', 'Africa/Freetown'), ('Africa/Lome', 'Africa/Lome'), ('Africa/Monrovia', 'Africa/Monrovia'), ('Africa/Nouakchott', 'Africa/Nouakchott'), ('Africa/Ouagadougou', 'Africa/Ouagadougou'), ('Africa/Sao_Tome', 'Africa/Sao_Tome'), ('Atlantic/Reykjavik', 'Atlantic/Reykjavik'), ('Atlantic/St_Helena', 'Atlantic/St_Helena'), ('Arctic/Longyearbyen', 'Arctic/Longyearbyen'), ('Europe/Amsterdam', 'Europe/Amsterdam'), ('Europe/Andorra', 'Europe/Andorra'), ('Europe/Berlin', 'Europe/Berlin'), ('Europe/Busingen', 'Europe/Busingen'), ('Europe/Gibraltar', 'Europe/Gibraltar'), ('Europe/Luxembourg', 'Europe/Luxembourg'), ('Europe/Malta', 'Europe/Malta'), ('Europe/Monaco', 'Europe/Monaco'), ('Europe/Oslo', 'Europe/Oslo'), ('Europe/Rome', 'Europe/Rome'), ('Europe/San_Marino', 'Europe/San_Marino'), ('Europe/Stockholm', 'Europe/Stockholm'), ('Europe/Vaduz', 'Europe/Vaduz'), ('Europe/Vatican', 'Europe/Vatican'), ('Europe/Vienna', 'Europe/Vienna'), ('Europe/Zurich', 'Europe/Zurich'), ('Europe/Belgrade', 'Europe/Belgrade'), ('Europe/Bratislava', 'Europe/Bratislava'), ('Europe/Budapest', 'Europe/Budapest'), ('Europe/Ljubljana', 'Europe/Ljubljana'), ('Europe/Podgorica', 'Europe/Podgorica'), ('Europe/Prague', 'Europe/Prague'), ('Europe/Tirane', 'Europe/Tirane'), ('Africa/Ceuta', 'Africa/Ceuta'), ('Europe/Brussels', 'Europe/Brussels'), ('Europe/Copenhagen', 'Europe/Copenhagen'), ('Europe/Madrid', 'Europe/Madrid'), ('Europe/Paris', 'Europe/Paris'), ('Europe/Sarajevo', 'Europe/Sarajevo'), ('Europe/Skopje', 'Europe/Skopje'), ('Europe/Warsaw', 'Europe/Warsaw'), ('Europe/Zagreb', 'Europe/Zagreb'), ('Africa/Algiers', 'Africa/Algiers'), ('Africa/Bangui', 'Africa/Bangui'), ('Africa/Brazzaville', 'Africa/Brazzaville'), ('Africa/Douala', 'Africa/Douala'), ('Africa/Kinshasa', 'Africa/Kinshasa'), ('Africa/Lagos', 'Africa/Lagos'), ('Africa/Libreville', 'Africa/Libreville'), ('Africa/Luanda', 'Africa/Luanda'), ('Africa/Malabo', 'Africa/Malabo'), ('Africa/Ndjamena', 'Africa/Ndjamena'), ('Africa/Niamey', 'Africa/Niamey'), ('Africa/Porto-Novo', 'Africa/Porto-Novo'), ('Africa/Tunis', 'Africa/Tunis'), ('Etc/GMT-1', 'Etc/GMT-1'), ('Africa/Windhoek', 'Africa/Windhoek'), ('Asia/Nicosia', 'Asia/Nicosia'), ('Europe/Athens', 'Europe/Athens'), ('Europe/Bucharest', 'Europe/Bucharest'), ('Europe/Chisinau', 'Europe/Chisinau'), ('Asia/Beirut', 'Asia/Beirut'), ('Africa/Cairo', 'Africa/Cairo'), ('Asia/Damascus', 'Asia/Damascus'), ('Europe/Helsinki', 'Europe/Helsinki'), ('Europe/Kiev', 'Europe/Kiev'), ('Europe/Mariehamn', 'Europe/Mariehamn'), ('Europe/Nicosia', 'Europe/Nicosia'), ('Europe/Riga', 'Europe/Riga'), ('Europe/Sofia', 'Europe/Sofia'), ('Europe/Tallinn', 'Europe/Tallinn'), ('Europe/Uzhgorod', 'Europe/Uzhgorod'), ('Europe/Vilnius', 'Europe/Vilnius'), ('Europe/Zaporozhye', 'Europe/Zaporozhye'), ('Africa/Blantyre', 'Africa/Blantyre'), ('Africa/Bujumbura', 'Africa/Bujumbura'), ('Africa/Gaborone', 'Africa/Gaborone'), ('Africa/Harare', 'Africa/Harare'), ('Africa/Johannesburg', 'Africa/Johannesburg'), ('Africa/Kigali', 'Africa/Kigali'), ('Africa/Lubumbashi', 'Africa/Lubumbashi'), ('Africa/Lusaka', 'Africa/Lusaka'), ('Africa/Maputo', 'Africa/Maputo'), ('Africa/Maseru', 'Africa/Maseru'), ('Africa/Mbabane', 'Africa/Mbabane'), ('Etc/GMT-2', 'Etc/GMT-2'), ('Europe/Istanbul', 'Europe/Istanbul'), ('Asia/Jerusalem', 'Asia/Jerusalem'), ('Africa/Tripoli', 'Africa/Tripoli'), ('Asia/Amman', 'Asia/Amman'), ('Asia/Baghdad', 'Asia/Baghdad'), ('Europe/Kaliningrad', 'Europe/Kaliningrad'), ('Europe/Minsk', 'Europe/Minsk'), ('Asia/Aden', 'Asia/Aden'), ('Asia/Bahrain', 'Asia/Bahrain'), ('Asia/Kuwait', 'Asia/Kuwait'), ('Asia/Qatar', 'Asia/Qatar'), ('Asia/Riyadh', 'Asia/Riyadh'), ('Africa/Addis_Ababa', 'Africa/Addis_Ababa'), ('Africa/Asmera', 'Africa/Asmera'), ('Africa/Dar_es_Salaam', 'Africa/Dar_es_Salaam'), ('Africa/Djibouti', 'Africa/Djibouti'), ('Africa/Juba', 'Africa/Juba'), ('Africa/Kampala', 'Africa/Kampala'), ('Africa/Khartoum', 'Africa/Khartoum'), ('Africa/Mogadishu', 'Africa/Mogadishu'), ('Africa/Nairobi', 'Africa/Nairobi'), ('Antarctica/Syowa', 'Antarctica/Syowa'), ('Etc/GMT-3', 'Etc/GMT-3'), ('Indian/Antananarivo', 'Indian/Antananarivo'), ('Indian/Comoro', 'Indian/Comoro'), ('Indian/Mayotte', 'Indian/Mayotte'), ('Europe/Kirov', 'Europe/Kirov'), ('Europe/Moscow', 'Europe/Moscow'), ('Europe/Simferopol', 'Europe/Simferopol'), ('Europe/Volgograd', 'Europe/Volgograd'), ('Europe/Astrakhan', 'Europe/Astrakhan'), ('Europe/Samara', 'Europe/Samara'), ('Europe/Ulyanovsk', 'Europe/Ulyanovsk'), ('Asia/Tehran', 'Asia/Tehran'), ('Asia/Dubai', 'Asia/Dubai'), ('Asia/Muscat', 'Asia/Muscat'), ('Etc/GMT-4', 'Etc/GMT-4'), ('Asia/Baku', 'Asia/Baku'), ('Indian/Mahe', 'Indian/Mahe'), ('Indian/Mauritius', 'Indian/Mauritius'), ('Indian/Reunion', 'Indian/Reunion'), ('Asia/Tbilisi', 'Asia/Tbilisi'), ('Asia/Yerevan', 'Asia/Yerevan'), ('Asia/Kabul', 'Asia/Kabul'), ('Antarctica/Mawson', 'Antarctica/Mawson'), ('Asia/Aqtau', 'Asia/Aqtau'), ('Asia/Aqtobe', 'Asia/Aqtobe'), ('Asia/Ashgabat', 'Asia/Ashgabat'), ('Asia/Dushanbe', 'Asia/Dushanbe'), ('Asia/Oral', 'Asia/Oral'), ('Asia/Samarkand', 'Asia/Samarkand'), ('Asia/Tashkent', 'Asia/Tashkent'), ('Etc/GMT-5', 'Etc/GMT-5'), ('Indian/Kerguelen', 'Indian/Kerguelen'), ('Indian/Maldives', 'Indian/Maldives'), ('Asia/Yekaterinburg', 'Asia/Yekaterinburg'), ('Asia/Karachi', 'Asia/Karachi'), ('Asia/Kolkata', 'Asia/Kolkata'), ('Asia/Colombo', 'Asia/Colombo'), ('Asia/Katmandu', 'Asia/Katmandu'), ('Antarctica/Vostok', 'Antarctica/Vostok'), ('Asia/Almaty', 'Asia/Almaty'), ('Asia/Bishkek', 'Asia/Bishkek'), ('Asia/Qyzylorda', 'Asia/Qyzylorda'), ('Asia/Urumqi', 'Asia/Urumqi'), ('Etc/GMT-6', 'Etc/GMT-6'), ('Indian/Chagos', 'Indian/Chagos'), ('Asia/Dhaka', 'Asia/Dhaka'), ('Asia/Thimphu', 'Asia/Thimphu'), ('Asia/Rangoon', 'Asia/Rangoon'), ('Indian/Cocos', 'Indian/Cocos'), ('Antarctica/Davis', 'Antarctica/Davis'), ('Asia/Bangkok', 'Asia/Bangkok'), ('Asia/Hovd', 'Asia/Hovd'), ('Asia/Jakarta', 'Asia/Jakarta'), ('Asia/Phnom_Penh', 'Asia/Phnom_Penh'), ('Asia/Pontianak', 'Asia/Pontianak'), ('Asia/Saigon', 'Asia/Saigon'), ('Asia/Vientiane', 'Asia/Vientiane'), ('Etc/GMT-7', 'Etc/GMT-7'), ('Indian/Christmas', 'Indian/Christmas'), ('Asia/Novokuznetsk', 'Asia/Novokuznetsk'), ('Asia/Novosibirsk', 'Asia/Novosibirsk'), ('Asia/Omsk', 'Asia/Omsk'), ('Asia/Hong_Kong', 'Asia/Hong_Kong'), ('Asia/Macau', 'Asia/Macau'), ('Asia/Shanghai', 'Asia/Shanghai'), ('Asia/Krasnoyarsk', 'Asia/Krasnoyarsk'), ('Asia/Brunei', 'Asia/Brunei'), ('Asia/Kuala_Lumpur', 'Asia/Kuala_Lumpur'), ('Asia/Kuching', 'Asia/Kuching'), ('Asia/Makassar', 'Asia/Makassar'), ('Asia/Manila', 'Asia/Manila'), ('Asia/Singapore', 'Asia/Singapore'), ('Etc/GMT-8', 'Etc/GMT-8'), ('Antarctica/Casey', 'Antarctica/Casey'), ('Australia/Perth', 'Australia/Perth'), ('Asia/Taipei', 'Asia/Taipei'), ('Asia/Choibalsan', 'Asia/Choibalsan'), ('Asia/Ulaanbaatar', 'Asia/Ulaanbaatar'), ('Asia/Irkutsk', 'Asia/Irkutsk'), ('Asia/Dili', 'Asia/Dili'), ('Asia/Jayapura', 'Asia/Jayapura'), ('Asia/Tokyo', 'Asia/Tokyo'), ('Etc/GMT-9', 'Etc/GMT-9'), ('Pacific/Palau', 'Pacific/Palau'), ('Asia/Pyongyang', 'Asia/Pyongyang'), ('Asia/Seoul', 'Asia/Seoul'), ('Australia/Adelaide', 'Australia/Adelaide'), ('Australia/Broken_Hill', 'Australia/Broken_Hill'), ('Australia/Darwin', 'Australia/Darwin'), ('Australia/Brisbane', 'Australia/Brisbane'), ('Australia/Lindeman', 'Australia/Lindeman'), ('Australia/Melbourne', 'Australia/Melbourne'), ('Australia/Sydney', 'Australia/Sydney'), ('Antarctica/DumontDUrville', 'Antarctica/DumontDUrville'), ('Etc/GMT-10', 'Etc/GMT-10'), ('Pacific/Guam', 'Pacific/Guam'), ('Pacific/Port_Moresby', 'Pacific/Port_Moresby'), ('Pacific/Saipan', 'Pacific/Saipan'), ('Pacific/Truk', 'Pacific/Truk'), ('Australia/Currie', 'Australia/Currie'), ('Australia/Hobart', 'Australia/Hobart'), ('Asia/Chita', 'Asia/Chita'), ('Asia/Khandyga', 'Asia/Khandyga'), ('Asia/Yakutsk', 'Asia/Yakutsk'), ('Antarctica/Macquarie', 'Antarctica/Macquarie'), ('Etc/GMT-11', 'Etc/GMT-11'), ('Pacific/Efate', 'Pacific/Efate'), ('Pacific/Guadalcanal', 'Pacific/Guadalcanal'), ('Pacific/Kosrae', 'Pacific/Kosrae'), ('Pacific/Noumea', 'Pacific/Noumea'), ('Pacific/Ponape', 'Pacific/Ponape'), ('Asia/Sakhalin', 'Asia/Sakhalin'), ('Asia/Ust-Nera', 'Asia/Ust-Nera'), ('Asia/Vladivostok', 'Asia/Vladivostok'), ('Antarctica/McMurdo', 'Antarctica/McMurdo'), ('Pacific/Auckland', 'Pacific/Auckland'), ('Etc/GMT-12', 'Etc/GMT-12'), ('Pacific/Funafuti', 'Pacific/Funafuti'), ('Pacific/Kwajalein', 'Pacific/Kwajalein'), ('Pacific/Majuro', 'Pacific/Majuro'), ('Pacific/Nauru', 'Pacific/Nauru'), ('Pacific/Tarawa', 'Pacific/Tarawa'), ('Pacific/Wake', 'Pacific/Wake'), ('Pacific/Wallis', 'Pacific/Wallis'), ('Pacific/Fiji', 'Pacific/Fiji'), ('Asia/Anadyr', 'Asia/Anadyr'), ('Asia/Kamchatka', 'Asia/Kamchatka'), ('Asia/Magadan', 'Asia/Magadan'), ('Asia/Srednekolymsk', 'Asia/Srednekolymsk'), ('Etc/GMT-13', 'Etc/GMT-13'), ('Pacific/Enderbury', 'Pacific/Enderbury'), ('Pacific/Fakaofo', 'Pacific/Fakaofo'), ('Pacific/Tongatapu', 'Pacific/Tongatapu'), ('Pacific/Apia', 'Pacific/Apia')], default=trionyx.trionyx.models.default_timezone, max_length=32, verbose_name='Timezone')),
                ('created_by', models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL, verbose_name='Created by')),
                ('groups', models.ManyToManyField(blank=True, help_text='The groups this user belongs to. A user will get all permissions granted to each of their groups.', related_name='user_set', related_query_name='user', to='auth.Group', verbose_name='groups')),
                ('user_permissions', models.ManyToManyField(blank=True, help_text='Specific permissions for this user.', related_name='user_set', related_query_name='user', to='auth.Permission', verbose_name='user permissions')),
            ],
            options={
                'verbose_name': 'User',
                'verbose_name_plural': 'Users',
            },
        ),
        migrations.CreateModel(
            name='Log',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='Created at')),
                ('updated_at', models.DateTimeField(auto_now=True, verbose_name='Updated at')),
                ('deleted', models.BooleanField(default=False, verbose_name='Deleted')),
                ('verbose_name', models.TextField(blank=True, default='', verbose_name='Verbose name')),
                ('log_hash', models.CharField(max_length=32, verbose_name='Log hash')),
                ('level', models.IntegerField(choices=[(50, 'Critical'), (40, 'Error'), (30, 'Warning'), (20, 'Info'), (10, 'Debug'), (0, 'Not set')], verbose_name='Level')),
                ('message', models.TextField(verbose_name='Message')),
                ('file_path', models.CharField(max_length=256, verbose_name='File path')),
                ('file_line', models.IntegerField(verbose_name='File line')),
                ('traceback', models.TextField(default='', verbose_name='Traceback')),
                ('last_event', models.DateTimeField(verbose_name='Last event')),
                ('log_count', models.IntegerField(default=1, verbose_name='Log count')),
                ('created_by', models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL, verbose_name='Created by')),
            ],
            options={
                'verbose_name': 'Log',
                'verbose_name_plural': 'Logs',
            },
        ),
        migrations.CreateModel(
            name='Task',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='Created at')),
                ('updated_at', models.DateTimeField(auto_now=True, verbose_name='Updated at')),
                ('deleted', models.BooleanField(default=False, verbose_name='Deleted')),
                ('verbose_name', models.TextField(blank=True, default='', verbose_name='Verbose name')),
                ('celery_task_id', models.CharField(max_length=64, unique=True)),
                ('status', models.IntegerField(choices=[(10, 'Scheduled'), (20, 'Queue'), (30, 'Locked'), (40, 'Running'), (50, 'Completed'), (99, 'Failed')], default=20, verbose_name='Status')),
                ('identifier', models.CharField(default='not_set', max_length=255, verbose_name='Identifier')),
                ('description', models.TextField(default='', verbose_name='Description')),
                ('progress', models.PositiveIntegerField(default=0, verbose_name='Progress')),
                ('progress_output', jsonfield.fields.JSONField(default=list, dump_kwargs={'cls': jsonfield.encoder.JSONEncoder, 'separators': (',', ':')}, load_kwargs={}, verbose_name='Progress output')),
                ('scheduled_at', models.DateTimeField(blank=True, null=True, verbose_name='Scheduled at')),
                ('started_at', models.DateTimeField(blank=True, null=True, verbose_name='started at')),
                ('execution_time', models.IntegerField(default=0, verbose_name='Execution time')),
                ('result', models.TextField(blank=True, default='', verbose_name='Result')),
                ('object_id', models.BigIntegerField(blank=True, null=True, verbose_name='Object ID')),
                ('object_verbose_name', models.TextField(blank=True, default='', verbose_name='Object name')),
                ('created_by', models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL, verbose_name='Created by')),
                ('object_type', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='contenttypes.ContentType', verbose_name='Object type')),
                ('user', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL, verbose_name='Started by')),
            ],
            options={
                'verbose_name': 'Task',
                'verbose_name_plural': 'Tasks',
            },
        ),
        migrations.CreateModel(
            name='LogEntry',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('log_time', models.DateTimeField(verbose_name='Log time')),
                ('path', models.TextField(default='', verbose_name='Path')),
                ('user_agent', models.TextField(default='', verbose_name='User agent')),
                ('log', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='entries', to='trionyx.Log')),
                ('user', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL, verbose_name='User')),
            ],
            options={
                'verbose_name': 'Log entry',
                'verbose_name_plural': 'Log entries',
            },
        ),
        migrations.CreateModel(
            name='AuditLogEntry',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='Created at')),
                ('updated_at', models.DateTimeField(auto_now=True, verbose_name='Updated at')),
                ('deleted', models.BooleanField(default=False, verbose_name='Deleted')),
                ('verbose_name', models.TextField(blank=True, default='', verbose_name='Verbose name')),
                ('object_id', models.BigIntegerField(blank=True, null=True)),
                ('object_verbose_name', models.TextField(blank=True, default='')),
                ('action', models.IntegerField(choices=[(10, 'Added'), (20, 'Changed'), (30, 'Deleted')])),
                ('changes', jsonfield.fields.JSONField(dump_kwargs={'cls': jsonfield.encoder.JSONEncoder, 'separators': (',', ':')}, load_kwargs={})),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='contenttypes.ContentType')),
                ('created_by', models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL, verbose_name='Created by')),
                ('user', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='UserAttribute',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('code', models.CharField(max_length=128)),
                ('value', jsonfield.fields.JSONField(dump_kwargs={'cls': jsonfield.encoder.JSONEncoder, 'separators': (',', ':')}, load_kwargs={})),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='attributes', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'unique_together': {('user', 'code')},
            },
        ),
        migrations.AddIndex(
            model_name='auditlogentry',
            index=models.Index(fields=['content_type', 'object_id'], name='trionyx_aud_content_072d99_idx'),
        ),
    ]



"""
trionyx.config
~~~~~~~~~~~~~~

:copyright: 2017 by Maikel Martens
:license: GPLv3
"""
import inspect
from functools import reduce
from typing import Optional, Generator, Union

from django.apps import apps
from django.urls import reverse
from django.conf import settings
from django.db.models import Field, Model

TX_MODEL_CONFIGS = settings.TX_CORE_MODEL_CONFIGS
TX_MODEL_CONFIGS.update(settings.TX_MODEL_CONFIGS)
TX_MODEL_OVERWRITES = {key.lower(): value.lower() for key, value in settings.TX_MODEL_OVERWRITES.items()}


class ModelConfig:
    """
    ModelConfig holds all config related to a model that is used for trionyx functionality.

    Model configs are auto loaded from the apps config file.
    In the apps config class create a class with same name as model and set appropriate config as class attribute.

    .. code-block:: python

        # apps.blog.apps.py
        class BlogConfig(BaseConfig):
            ...

            # Example config for Category model
            class Category:
                verbose_name = '{name}'
                list_default_fields = ['id', 'created_at', 'name']
                list_search_fields = ['name', 'description']

    """

    menu_name = None
    """Menu name, default is model verbose_name_plural"""

    menu_order = None
    """Menu order"""

    menu_exclude = False
    """Exclude model from menu"""

    menu_root = False
    """Add menu item to root instead of under the app menu"""

    menu_icon = None
    """Menu css icon, is ony used when root menu item"""

    global_search = True
    """Enable global search for model"""

    disable_search_index = False
    """Disable search index, use full for model with no list view but with allot of records"""

    search_fields = ()
    """Fields to use for searching, default is all CharField and TextField"""

    search_exclude_fields = ()
    """Fields you don't want to use for search"""

    search_title = None
    """
    Search title of model works the same as `verbose_name`, defaults to __str__.
    Is given high priority in search and is used in global search
    """

    search_description = None
    """
    Search description of model works the same as `verbose_name`, default is empty,
    Is given medium priority and is used in global search page
    """

    list_fields = None
    """
    Customise the available fields for model list view, default all model fields are available.

    list_fields is an array of dict with the field description, the following options are available:

    - **field**: Model field name (is used for sort and getting value if no renderer is supplied)
    - **label**: Column name in list view, if not set verbose_name of model field is used
    - **renderer**: function(model, field) that returns a JSON serializable date, when not set model field is used.

    .. code-block:: python

         list_fields = [
            {
                'field': 'first_name',
                'label': 'Real first name',
                'renderer': lambda model field: model.first_name.upper()
            }
         ]
    """

    list_default_fields = None
    """Array of fields that default is used in form list"""

    list_select_related = None
    """Array of fields to add foreign-key relationships to query, use this for relations that are used in search or renderer"""

    list_default_sort = '-id'
    """Default sort field for list view"""

    api_fields = None
    """Fields used in API model serializer, fallback on fields used in create and edit forms"""

    api_disable = False
    """Disable API for model"""

    verbose_name = "{model_name}({id})"
    """
    Verbose name used for displaying model, default value is "{model_name}({id})"

    format can be used to get model attributes value, there are two extra values supplied:
        - app_label: App name
        - model_name: Class name of model
    """

    view_header_buttons = None
    """
    List with button configurations to be displayed in view header bar

    .. code-block:: python

         view_header_buttons = [
            {
                'label': 'Send email', # string or function
                'url': lambda obj : reverse('blog.post', kwargs={'pk': obj.id}), # string or function
                'type': 'default',
                'show': lambda obj, alias : True, # Function that gives True or False if button must be displayed
                'modal': True,
            }
         ]
    """

    disable_add = False
    """Disable add for this model"""

    disable_change = False
    """Disable change for this model"""

    disable_delete = False
    """Disable delete for this model"""

    auditlog_disable = False
    """Disable auditlog for this model"""

    auditlog_ignore_fields = None
    """Auditlog fields to be ignored"""

    hide_permissions = False
    """Dont show model in permissions tree, prevent clutter from internal models"""

    def __init__(self, model: Model, MetaConfig=None):
        """Init config"""
        self.model = model
        self.app_config = apps.get_app_config(model._meta.app_label)
        self.app_label = model._meta.app_label
        self.model_name = model._meta.model_name
        self.__changed = {}

        if MetaConfig:
            for key, value in MetaConfig.__dict__.items():
                if key.startswith('_'):
                    continue
                setattr(self, key, value)

    def __setattr__(self, name, value):
        """Add attribute to changed list"""
        try:
            self.__changed[name] = True
        except Exception:
            # Ignore errors from __init__ that __changed does not exists
            pass
        super().__setattr__(name, value)

    def __getattr__(self, item):
        """Get attribute and returns null if not set"""
        try:
            return super().__getattr__(item)
        except AttributeError:
            return None

    def get_app_verbose_name(self, title: bool = True) -> str:
        """Get app verbose name"""
        return str(self.app_config.verbose_name).title() if title else str(self.app_config.verbose_name)

    def get_verbose_name(self, title: bool = True) -> str:
        """Get class verbose name"""
        return str(
            self.model._meta.verbose_name
        ).title() if title else str(self.model._meta.verbose_name).lower()

    def get_verbose_name_plural(self, title: bool = True) -> str:
        """Get class plural verbose name"""
        return str(
            self.model._meta.verbose_name_plural
        ).title() if title else str(self.model._meta.verbose_name_plural).lower()

    @property
    def is_trionyx_model(self) -> bool:
        """Check if config is for Trionyx model"""
        from trionyx.models import BaseModel
        return isinstance(self.model(), BaseModel)

    def has_config(self, name: str) -> bool:
        """Check if config is set"""
        return name in self.__changed

    def get_field(self, field_name):
        """Get model field by name"""
        return self.model._meta.get_field(field_name)

    def get_fields(self, inlcude_base: bool = False, include_id: bool = False):
        """Get model fields"""
        for field in self.model._meta.fields:
            if field.name == 'deleted':
                continue
            if not include_id and field.name == 'id':
                continue
            if not inlcude_base and field.name in ['created_at', 'updated_at', 'created_by', 'verbose_name']:
                continue
            yield field

    def get_url(self, view_name: str, model: Model = None, code: str = None) -> str:
        """Get url for model"""
        from trionyx.urls import model_url
        return model_url(
            model if model else self.model,
            view_name,
            code
        )

    def get_absolute_url(self, model: Model) -> str:
        """Get model url"""
        return reverse('trionyx:model-view', kwargs={
            'app': model._meta.app_label,
            'model': model._meta.model_name,
            'pk': model.pk
        })

    def get_list_fields(self) -> [dict]:
        """Get all list fields"""
        from trionyx.renderer import renderer
        model_fields = {f.name: f for f in self.get_fields(True, True)}

        def create_list_fields(config_fields, list_fields=None):
            list_fields = list_fields if list_fields else {}

            for field_config in config_fields:
                config = field_config
                if isinstance(field_config, str):
                    config = {'field': field_config}

                if 'field' not in config:
                    raise Exception("Field config is missing field: {}".format(config))

                field_model_config = self
                field_parts = config['field'].split('__')

                if len(field_parts) > 1:
                    related_class = reduce(
                        lambda obj, field: getattr(obj, field).field.related_model,
                        field_parts[:-1],
                        self.model)

                    field_model_config = models_config.get_config(related_class)
                    field = field_model_config.get_field(field_parts[-1])

                    config['choices_url'] = self.get_url('list-choices')
                elif config['field'] in model_fields:
                    field = model_fields[config['field']]
                else:
                    field = None

                if 'label' not in config:
                    label = field.verbose_name if field else config['field']
                    label = '{} {}'.format(field_model_config.get_verbose_name(), label) if len(field_parts) > 1 else label
                    config['label'] = label

                if 'renderer' not in config:
                    config['renderer'] = renderer.render_field

                if 'type' not in config and field:
                    config['type'] = self.get_field_type(field)

                if 'choices' not in config and field:
                    config['choices'] = field.choices

                list_fields[config['field']] = config
            return list_fields

        list_fields = create_list_fields(model_fields.keys())

        if self.list_fields:
            list_fields = create_list_fields(self.list_fields, list_fields)

        return list_fields

    def get_field_type(self, field: Field) -> str:
        """Get field type base on model field class"""
        from trionyx import models
        if isinstance(field, (models.CharField, models.TextField, models.EmailField)):
            return 'text'
        elif isinstance(field, (models.IntegerField, models.AutoField)):
            return 'int'
        elif isinstance(field, (models.DecimalField, models.FloatField)):
            return 'float'
        elif isinstance(field, models.BooleanField):
            return 'bool'
        elif isinstance(field, models.DateTimeField):
            return 'datetime'
        elif isinstance(field, models.DateField):
            return 'date'
        elif isinstance(field, (models.ForeignKey, models.OneToOneField)):
            return 'related'


class Models:
    """Holds all model configs"""

    def __init__(self):
        """Init models"""
        self.configs = {}

    def auto_load_configs(self):
        """Auto load all configs from app configs"""
        from trionyx.models import BaseModel

        for app in apps.get_app_configs():
            for model in app.get_models():
                config = ModelConfig(model, getattr(app, model.__name__, None))
                if not isinstance(config.model(), BaseModel):
                    config.disable_search_index = True
                self.configs[self.get_model_name(model)] = config

        # Merge not set configs from overwrites
        for old, new in TX_MODEL_OVERWRITES.items():
            old_config = self.configs[old]
            new_config = self.configs[new]

            for key, value in old_config.__dict__.items():
                if (
                    key.startswith('_')
                    or key in ['model', 'app_config', 'app_label', 'model_name']
                    or new_config.has_config(key)
                ):
                    continue
                setattr(new_config, key, value)

        # Update configs from settings
        for model_name, config in TX_MODEL_CONFIGS.items():
            for key, value in config.items():
                setattr(self.configs[model_name], key, value)

    def get_config(self, model: Union[str, Model]) -> Optional[ModelConfig]:
        """Get config for given model"""
        if not inspect.isclass(model) and not isinstance(model, str):
            model = model.__class__
        return self.configs.get(self.get_model_name(model))

    def get_all_configs(self, trionyx_models_only: bool = True) -> Generator[ModelConfig, None, None]:
        """Get all model configs"""
        from trionyx.models import BaseModel

        for index, config in self.configs.items():
            if trionyx_models_only and not isinstance(config.model(), BaseModel):
                continue

            yield config

    def get_model_name(self, model: Union[str, Model]) -> str:
        """Get model name for given model"""
        if isinstance(model, str):
            return model.lower()
        return '{}.{}'.format(model._meta.app_label, model._meta.model_name).lower()

    def get_all_models(self, user: Optional["trionyx.trionyx.models.User"] = None, trionyx_models_only: bool = True):
        """Get all user models"""
        for config in self.get_all_configs(trionyx_models_only):
            if config.app_label == 'trionyx' and config.model_name in ['session', 'auditlogentry', 'log', 'logentry', 'userattribute']:
                continue

            if user and user.has_perm('{app_label}.view_{model_name}'.format(
                app_label=config.app_label,
                model_name=config.model_name,
            ).lower()):
                yield config.model
            elif not user:
                yield config.model


models_config = Models()



"""
trionyx.trionyx.api.pagination
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: 2019 by Maikel Martens
:license: GPLv3
"""
from collections import OrderedDict
from rest_framework.response import Response

from rest_framework.pagination import PageNumberPagination


class PageNumberPagination(PageNumberPagination):
    """Api Pagination class"""

    def get_paginated_response(self, data):
        """Get paginated response, added extra fields"""
        return Response(OrderedDict([
            ('count', self.page.paginator.count),
            ('page', self.page.number),
            ('per_page', self.page.paginator.per_page),
            ('num_pages', self.page.paginator.num_pages),
            ('next', self.get_next_link()),
            ('previous', self.get_previous_link()),
            ('results', data)
        ]))



"""
trionyx.trionyx.management.commands.create_app
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: 2018 by Maikel Martens
:license: GPLv3
"""
import os

from django.core.management.base import BaseCommand, CommandError
from django.conf import settings

from trionyx.quickstart import Quickstart


class Command(BaseCommand):
    """Command for creating base Trionyx app"""

    help = 'Create base app'

    def add_arguments(self, parser):
        """Add name argument"""
        parser.add_argument('name', type=str)

    def handle(self, *args, **options):
        """Create new app"""
        quickstart = Quickstart()

        try:
            quickstart.create_app(os.path.join(settings.BASE_DIR, 'apps'), options.get('name'))
            self.stdout.write(
                self.style.SUCCESS("Successfully created app ({name}), don't forget to add 'apps.{name}' to INSTALLED_APPS".format(
                    name=options.get('name')
                ))
            )
        except FileExistsError as e:
            print(e)
            raise CommandError("App with same name already exists")



"""
trionyx.log
~~~~~~~~~~~

Core app for Trionyx

:copyright: 2019 by Maikel Martens
:license: GPLv3
"""
import logging
from django.conf import settings


class LogDBHandler(logging.Handler):
    """DB log handler"""

    def emit(self, record):
        """Save log record"""
        try:
            from trionyx.trionyx.models import Log
            Log.objects.create_log_entry_by_record(record)
        except Exception:
            pass  # Logception?


def enable_db_logger():
    """Enable DB logger"""
    db_handler = LogDBHandler()
    db_handler.setLevel(settings.TX_DB_LOG_LEVEL)
    logging.getLogger().addHandler(db_handler)



"""
trionyx.trionyx.management.commands.create_app
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: 2018 by Maikel Martens
:license: GPLv3
"""
from django.core.management.base import BaseCommand


class Command(BaseCommand):
    """Command for creating base Trionyx app"""

    help = 'Generate theme color fixes'

    def handle(self, *args, **options):
        """Create CSS"""
        skins = {
            'blue': '#0073b7',
            'yellow': '#f39c12',
            'green': '#00a65a',
            'purple': '#605ca8',
            'red': '#dd4b39',
            'black': '#111111',
        }

        css_color_selectors = {
            'background': {
                'css': [
                    'background-color: {color} !important;',
                    'color: #ffffff !important;'
                ],
                'selectors': [
                    '.bg-theme',
                    '.nav-pills > li.active > a',
                    '.pagination > li.active > a',
                    '.select2-container--default .select2-results__option--highlighted[aria-selected]',
                    '.select2-container--default .select2-selection--multiple .select2-selection__choice',
                    '.bootstrap-datetimepicker-widget table td.active',
                    '.box.box-solid.box-theme > .box-header',
                    '.progress-bar-theme',
                ]
            },
            'border': {
                'css': [
                    'border-color: {color} !important;'
                ],
                'selectors': [
                    '.select2-container--default.select2-container--focus .select2-selection--multiple',
                    '.select2-container--default .select2-search--dropdown .select2-search__field',
                    '.select2-container--default .select2-selection--multiple .select2-selection__choice',
                    '.nav-pills > li.active > a',
                    '.pagination > li.active > a',
                    '.box-theme',
                    '.form-control:focus',
                ]
            },
            'text': {
                'css': [
                    'color: {color} !important;',
                ],
                'selectors': [
                    '.text-theme',
                    '.bootstrap-datetimepicker-widget a',
                ],
            }
        }

        for skin, color in skins.items():
            for _, config in css_color_selectors.items():
                print(',\n'.join([
                    *['.skin-{} {}'.format(skin, s) for s in config['selectors']],
                    *['.skin-{}-light {}'.format(skin, s) for s in config['selectors']],
                ]))
                print('{')
                for css in config['css']:
                    print('    ' + css.format(color=color))
                print('}\n')



"""
trionyx.widgets
~~~~~~~~~~~~~~~

:copyright: 2019 by Maikel Martens
:license: GPLv3
"""
import json

from django.utils import timezone
from django.contrib.staticfiles.templatetags.staticfiles import static
from django.contrib.contenttypes.models import ContentType
from trionyx.trionyx.models import AuditLogEntry
from trionyx.renderer import renderer
from trionyx.config import models_config
from trionyx.trionyx.forms import AuditlogWidgetForm, TotalSummaryWidgetForm
from trionyx.models import Sum, filter_queryset_with_user_filters
from django.utils.translation import ugettext_lazy as _


widgets = {}


class MetaClass(type):
    """MetaClass for widget"""

    def __new__(cls, clsname, bases, attrs):
        """Auto register new widget cass"""
        newclass = super().__new__(cls, clsname, bases, attrs)
        if clsname != 'BaseWidget':
            widgets[newclass.code] = newclass
        return newclass


class BaseWidget(metaclass=MetaClass):
    """
    Base widget to extend for creating custom widgets.
    Custom widgets are created in `widgets.py` in root of app folder.

    **Example of random widget:**

    .. code-block:: python

        # <app dir>/widgets.py
        RandomWidget(BaseWidget):
            code = 'random'
            name = 'Random widget'
            description = 'Shows random string'

            def get_data(self, request, config):
                return utils.random_string(16)


    .. code-block:: html

        <!-- template path: widgets/random.html -->
        <script type="text/x-template" id="widget-random-template">
            <div :class="widgetClass">
                <div class="box-header with-border">
                    <!-- Get title from config, your form fields are also available in the config -->
                    <h3 class="box-title">[[widget.config.title]]</h3>
                </div>
                <!-- /.box-header -->
                <div class="box-body">
                    <!-- vue data property will be filled with the get_data results method --->
                    [[data]]
                </div>
              </div>
        </script>


        <script>
            <!-- The component must be called `widget-<code>` -->
            Vue.component('widget-random', {
                mixins: [TxWidgetMixin],
                template: '#widget-random-template',
            });
        </script>

    """

    code = None
    """Code for widget"""

    name = None
    """Name for widget is also used as default title"""

    description = None
    """Short description on what the widget does"""

    config_form_class = None
    """Form class used to change the widget. The form cleaned_data is used as the config"""

    default_width = 4
    """Default width of widget, is based on grid system with max 12 columns"""

    default_height = 20
    """Default height of widget, each step is 10px"""

    @property
    def template(self):
        """Template path `widgets/{code}.html` overwrite to set custom path"""
        return 'widgets/{code}.html'.format(code=self.code)

    @property
    def image(self):
        """Image path `img/widgets/{code}.jpg` overwrite to set custom path"""
        return 'img/widgets/{code}.jpg'.format(code=self.code)

    def get_data(self, request, config):
        """Get data for widget, function needs te be overwritten on widget implementation"""
        return None

    @property
    def config_fields(self):
        """Get the config field names"""
        if not self.config_form_class:
            return []

        fields = list(self.config_form_class().base_fields)

        for field in list(self.config_form_class().declared_fields):
            if field not in fields:
                fields.append(field)
        return fields


class AuditlogWidget(BaseWidget):
    """Auditlog widget"""

    code = 'auditlog'
    name = _('Latest actions')
    description = _('Show the latest tracked actions done by users and the system')
    config_form_class = AuditlogWidgetForm
    default_height = 22

    def get_data(self, request, config):
        """Get data for widget"""
        content_type_ids = [
            content_type.id for model, content_type
            in ContentType.objects.get_for_models(*[
                config.model for config in models_config.get_all_configs(False)
                if request.user.has_perm('{app_label}.view_{model_name}'.format(
                    app_label=config.app_label,
                    model_name=config.model_name,
                ).lower())
            ]).items()]

        logs = AuditLogEntry.objects.filter(content_type__in=content_type_ids).prefetch_related('user')
        show = config.get('show', 'all')
        if show != 'all':
            logs = logs.filter(user__isnull=show == 'system')

        return [
            {
                'user_full_name': log.user.get_full_name() if log.user else _('System'),
                'user_avatar': log.user.avatar.url if log.user and log.user.avatar else static('img/avatar.png'),
                'action': renderer.render_field(log, 'action'),
                'object': '({}) {}'.format(
                    log.content_type.model_class()._meta.verbose_name.capitalize(),
                    log.object_verbose_name),
                'object_url': log.content_object.get_absolute_url() if log.content_object else '',
                'created_at': renderer.render_field(log, 'created_at'),

            } for log in logs.order_by('-created_at')[:6]
        ]


class TotalSummaryWidget(BaseWidget):
    """Total summary widget"""

    code = 'total_summary'
    name = _('Total summary')
    description = _('Show total for given field on given period')
    config_form_class = TotalSummaryWidgetForm
    default_height = 5

    def get_data(self, request, config):
        """Get data"""
        ModelClass = ContentType.objects.get_for_id(config['model']).model_class()
        query = ModelClass.objects.get_queryset()

        if config.get('filters'):
            query = filter_queryset_with_user_filters(query, json.loads(config['filters']))

        if config.get('period', 'all') != 'all':
            today = timezone.now().replace(hour=0, minute=0, second=0, microsecond=0)
            query = query.filter(**{
                'year': {
                    '{}__gte'.format(config['period_field']): today.replace(month=1, day=1)
                },
                'month': {
                    '{}__gte'.format(config['period_field']): today.replace(day=1)
                },
                'week': {
                    '{}__gte'.format(config['period_field']): today - timezone.timedelta(today.weekday())
                },
                'day': {
                    '{}__gte'.format(config['period_field']): today
                },
                '365days': {
                    '{}__gte'.format(config['period_field']): today - timezone.timedelta(days=365)
                },
                '30days': {
                    '{}__gte'.format(config['period_field']): today - timezone.timedelta(days=30)
                },
                '7days': {
                    '{}__gte'.format(config['period_field']): today - timezone.timedelta(days=7)
                },
            }.get(config['period'], {}))

        if config.get('field', '__count__') == '__count__':
            return query.count()
        else:
            result = query.aggregate(sum=Sum(config['field']))
            return renderer.render_field(ModelClass(**{config['field']: result['sum']}), config['field'])



"""
trionyx.views
~~~~~~~~~~~~~

:copyright: 2018 by Maikel Martens
:license: GPLv3
"""
import inspect
import logging
from collections import defaultdict

from trionyx.config import models_config, TX_MODEL_OVERWRITES
from trionyx.layout import Layout, Column12, Panel, DescriptionList, Component
from django.utils.translation import ugettext_lazy as _

from .models import (  # noqa F401
    ListView, ListJsendView, ListExportView, ListChoicesJsendView, DetailTabView,
    DetailTabJsendView, UpdateView, CreateView, DeleteView, LayoutView,
)
from .dialogs import (  # noqa F401
    DialogView, UpdateDialog, CreateDialog, LayoutDialog, DeleteDialog
)

logger = logging.getLogger(__name__)


class LayoutRegister:
    """Class where tab layout can be registered"""

    def __init__(self):
        """Init"""
        self.layouts = {}

    def register(self, code):
        """Add layout to register"""
        def wrapper(create_layout):
            self.layouts[code] = create_layout
            return create_layout
        return wrapper

    def get_layout(self, code, object, request=None):
        """Get complete layout for given object"""
        if code not in self.layouts:
            raise Exception('layout does not exist')
        layout = self.layouts.get(code)
        layout = layout(object)
        if isinstance(layout, Component):
            layout = Layout(layout)

        if isinstance(layout, list):
            layout = Layout(*layout)

        layout.set_object(object)
        return layout.render(request)


class TabRegister:
    """Class where tab layout can be registered"""

    def __init__(self):
        """Init Tabs"""
        self.tabs = defaultdict(list)

    def get_tabs(self, model_alias, object):
        """
        Get all active tabs for given model

        :param model_alias:
        :param object: Object used to filter tabs
        :return:
        """
        model_alias = self.get_model_alias(model_alias)

        for item in self.tabs.get(model_alias, list()):
            if item.display_filter(object):
                yield item

    def get_tab(self, model_alias, object, tab_code):
        """
        Get tab for given object and tab code

        :param model_alias:
        :param object: Object used to render tab
        :param tab_code: Tab code to use
        :return:
        """
        model_alias = self.get_model_alias(model_alias)

        for item in self.tabs[model_alias]:
            if item.code == tab_code and item.display_filter(object):
                return item
        raise Exception('Given tab does not exits or is filtered')

    def register(self, model_alias, code='general', name=None, order=10, display_filter=None):
        """
        Register new tab

        :param model_alias:
        :param code:
        :param name:
        :param order:
        :return:
        """
        model_alias = self.get_model_alias(model_alias)

        if code == 'general' and not name:
            name = _('General')

        def wrapper(create_layout):
            item = TabItem(
                code=code,
                create_layout=create_layout,
                name=name,
                order=order,
                display_filter=display_filter
            )

            if item in self.tabs[model_alias]:
                raise Exception("Tab {} already registered for model {}".format(code, model_alias))

            self.tabs[model_alias].append(item)
            self.tabs[model_alias] = sorted(self.tabs[model_alias], key=lambda item: item.order if item.order else 10)

            return create_layout
        return wrapper

    def register_update(self, model_alias, code):
        """
        Register tab update function, function is being called with (layout, object)

        :param model_alias:
        :param code:
        :return:
        """
        model_alias = self.get_model_alias(model_alias)

        def wrapper(update_layout):
            for item in self.tabs[model_alias]:
                if item.code == code:
                    item.layout_updates.append(update_layout)
            return update_layout
        return wrapper

    def update(self, model_alias, code='general', name=None, order=10, display_filter=None):
        """
        Update given tab

        :param model_alias:
        :param code:
        :param name:
        :param order:
        :param display_filter:
        :return:
        """
        model_alias = self.get_model_alias(model_alias)
        for item in self.tabs[model_alias]:
            if item.code != code:
                continue
            if name:
                item.name = name
            if order:
                item.order = order
            if display_filter:
                item.display_filter = display_filter
            break
        self.tabs[model_alias] = sorted(self.tabs[model_alias], key=lambda item: item.code if item.code else 10)

    def get_model_alias(self, model_alias):
        """Get model alias if class then convert to alias string"""
        from trionyx.models import Model
        if inspect.isclass(model_alias) and issubclass(model_alias, Model):
            config = models_config.get_config(model_alias)
            name = '{}.{}'.format(config.app_label, config.model_name).lower()
        else:
            name = model_alias.lower()
        return TX_MODEL_OVERWRITES.get(name, name)

    def auto_generate_missing_tabs(self):
        """Auto generate tabs for models with no tabs"""
        for config in models_config.get_all_configs(False):
            model_alias = self.get_model_alias(config.model)
            if model_alias not in self.tabs:
                @self.register(model_alias, order=10)
                def general_layout(obj):
                    return Layout(
                        Column12(
                            Panel(
                                'info',
                                DescriptionList(*[f.name for f in models_config.get_config(obj).get_fields()])
                            )
                        )
                    )


class TabItem:
    """Tab item that holds the tab data and renders the layout"""

    def __init__(self, code, create_layout, name=None, order=None, display_filter=None):
        """Init TabItem"""
        self._name = None
        self.code = code
        self.create_layout = create_layout
        self.name = name
        self.order = order
        self.display_filter = display_filter if display_filter else lambda object: True
        self.layout_updates = []

    @property
    def name(self):
        """Give back tab name if is set else generate name by code"""
        if self._name:
            return self._name
        return self.code.replace('_', ' ').capitalize()

    @name.setter
    def name(self, name):
        """Set name"""
        self._name = name

    def get_layout(self, object):
        """Get complete layout for given object"""
        layout = self.create_layout(object)
        if isinstance(layout, Component):
            layout = Layout(layout)

        if isinstance(layout, list):
            layout = Layout(*layout)

        for update_layout in self.layout_updates:
            try:
                update_layout(layout, object)
            except Exception as e:
                logger.error('Could not update layout for {}: {}'.format(
                    self.code,
                    str(e)
                ))
        layout.set_object(object)
        return layout

    def __str__(self):
        """Tab string representation"""
        if not self.name:
            return self.name.capitalize()
        return self.name

    def __eq__(self, other):
        """Compare tab based on code"""
        return self.code == other.code


tabs = TabRegister()
layouts = LayoutRegister()



from trionyx import forms

from app.testblog.models import Tag


@forms.register(code='new', default_edit=True)
class TagForm(forms.ModelForm):
    name = forms.CharField(label='name (default edit form)')

    class Meta:
        model = Tag
        fields=['name']



"""
trionyx.trionyx.api.routers
~~~~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: 2019 by Maikel Martens
:license: GPLv3
"""
from importlib import import_module

from django.apps import apps
from rest_framework import routers
from rest_framework import serializers
from rest_framework.permissions import DjangoModelPermissions
from rest_framework.viewsets import ModelViewSet

from trionyx.config import models_config
from trionyx.forms import form_register
from trionyx.api.serializers import serializer_register


class AutoRouter(routers.DefaultRouter):
    """Auto router for adding all models to REST API"""

    @staticmethod
    def get_model_prefix(model):
        """Get model API prefix"""
        app_label = model._meta.app_label.replace(' ', '_').lower()
        model_name = model.__name__.lower()
        return '{}/{}'.format(app_label, model_name)

    def __init__(self):
        """Init class and find all model routes"""
        super(AutoRouter, self).__init__()
        self.auto_regiser_models()
        self.register_apps()

    def register_apps(self):
        """Register all API views from apps that are exported to routes.py"""
        for app in apps.get_app_configs():
            try:
                routes = import_module('{}.routes'.format(app.module.__package__))
            except (ImportError, TypeError):
                try:  # Fallback mode
                    routes = import_module('{}.routes'.format('.'.join(app.module.__package__.split('.')[:-2])))
                except (ImportError, TypeError):
                    continue

            if hasattr(routes, 'apiroutes'):
                for route in getattr(routes, 'apiroutes'):
                    self.register(**route)

    def auto_regiser_models(self):
        """Auto register all models"""
        # Load serializers
        for app in apps.get_app_configs():
            for module in ['serializers', 'api.serializers']:
                try:
                    import_module('{}.{}'.format(app.module.__package__, module))
                except ImportError:
                    pass

        for config in models_config.get_all_configs():
            if config.api_disable:
                continue

            model = config.model
            basename = model._meta.object_name.lower()
            classname = model.__name__
            serializer = serializer_register.get(model)

            DynamicViewSet = type(
                classname,
                (ModelViewSet,),
                {}
            )
            DynamicViewSet.model = model
            DynamicViewSet.queryset = model.objects.get_queryset()
            DynamicViewSet.serializer_class = serializer if serializer else self.generate_model_serializer(model, config)
            DynamicViewSet.permission_classes = (ExtendedDjangoModelPermissions,)
            DynamicViewSet.ordering = ['pk']

            # TODO maybe add serializer switcher https://www.django-rest-framework.org/api-guide/generic-views/#get_serializer_classself
            # So all API endpoint are read only but you can only see id and verbose name

            self.register(self.get_model_prefix(model), DynamicViewSet, basename)

    def generate_model_serializer(self, model, config):
        """Generate a model serializer with all fields"""
        class MetaModelSerializer(serializers.ModelSerializer):
            """Meta model serializer to make all fields work"""

            class Meta:
                model = None
                fields = None

        fields = []
        if config.api_fields:
            fields = config.api_fields
        else:
            forms = [
                form_register.get_create_form(model),
                form_register.get_edit_form(model)
            ]
            for form in forms:
                fields.extend([name for name in form.base_fields])

        fields = list({'id', 'created_at', 'updated_at', 'verbose_name', *fields})
        model_fields = [field.name for field in config.get_fields(True, True)]
        MetaModelSerializer.Meta.fields = [field for field in fields if field in model_fields]

        DynamicSerializer = type(
            '{}Serializer'.format(model.__name__),
            (MetaModelSerializer,),
            {}
        )
        DynamicSerializer.Meta.model = model

        return DynamicSerializer


class ExtendedDjangoModelPermissions(DjangoModelPermissions):
    """Permission class that added read as permission"""

    perms_map = {
        'GET': ['%(app_label)s.view_%(model_name)s'],
        'OPTIONS': [],
        'HEAD': [],
        'POST': ['%(app_label)s.add_%(model_name)s'],
        'PUT': ['%(app_label)s.change_%(model_name)s'],
        'PATCH': ['%(app_label)s.change_%(model_name)s'],
    }


def router(prefix, viewset, base_name):
    """
    Define an API route

    :param prefix:
    :param viewset:
    :param base_name:
    :return:
    """
    return {
        'prefix': prefix,
        'viewset': viewset,
        'base_name': base_name
    }



"""
trionyx.trionyx.tasks
~~~~~~~~~~~~~~~~~~~~~

:copyright: 2019 by Maikel Martens
:license: GPLv3
"""
import json
import math

from django.conf import settings
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _

from trionyx.trionyx.models import Task
from trionyx.tasks import shared_task, BaseTask
from trionyx.models import filter_queryset_with_user_filters


@shared_task
def cleanup_unexpectedly_stopped_tasks():
    """Set tasks that are unexpectedly stopped to failed"""
    Task.objects.filter(
        started_at__lt=timezone.now() - timezone.timedelta(seconds=settings.CELERY_TASK_TIME_LIMIT + 60),
        status__in=[Task.RUNNING, Task.LOCKED]
    ).update(
        status=Task.FAILED,
        result=_("Task unexpectedly stopped")
    )


class MassUpdateTask(BaseTask):
    """Mass update task"""

    name = 'mass_update'

    def run(self, all, ids, filters, data):
        """Run mass update"""
        query = self.get_model().objects.get_queryset()

        if all == '1':
            query = filter_queryset_with_user_filters(query, json.loads(filters))
        else:
            query = query.filter(id__in=[int(id) for id in filter(None, ids.split(','))])

        count = query.count()
        errors = []
        for index, obj in enumerate(query.iterator(100)):
            for key, value in data.items():
                setattr(obj, key, value)

            try:
                obj.save()
            except Exception:
                errors.append(str(obj))

            self.set_progress(math.ceil((index / count) * 100))

        if errors:
            raise Exception(_('Could not update the following items {}').format(', '.join(errors)))



"""
trionyx.trionyx.apps
~~~~~~~~~~~~~~~~~~~~

Core apps package containing Appconfig

:copyright: 2018 by Maikel Martens
:license: GPLv3
"""
from importlib import import_module

from django.apps import AppConfig
from django.apps import apps
from django.utils import timezone

from trionyx.config import models_config, ModelConfig
from trionyx.menu import app_menu
from trionyx.trionyx.search import auto_register_search_models
from trionyx.log import enable_db_logger
from django.utils.translation import ugettext_lazy as _
from django.contrib.staticfiles.templatetags.staticfiles import static

from .renderers import render_level, render_progress, render_status


class BaseConfig(AppConfig):
    """Base app config"""

    def get_model_config(self, model) -> ModelConfig:
        """Get model config for given model"""
        return models_config.get_config(model)


class Config(BaseConfig):
    """Trionyx core config app"""

    name = 'trionyx.trionyx'
    verbose_name = _('System')

    no_menu = True

    def ready(self):
        """Auto load Trionyx"""
        enable_db_logger()

        models_config.auto_load_configs()

        from trionyx import widgets  # noqa F401
        self.auto_load_app_modules(['layouts', 'signals', 'forms', 'widgets'])

        app_menu.auto_load_model_menu()

        auto_register_search_models()

        from trionyx.views import tabs
        tabs.auto_generate_missing_tabs()

        from trionyx.trionyx.auditlog import init_auditlog
        init_auditlog()

        # Add admin menu items
        from trionyx.urls import model_url
        app_menu.add_item('dashboard', _('Dashboard'), url='/', icon='fa fa-dashboard', order=1)
        app_menu.add_item('admin', _('Admin'), icon='fa fa-cogs', order=9000, permission='is_superuser')
        app_menu.add_item('admin/users', _('Users'), url=model_url('trionyx.user', 'list'), order=9010, permission='is_superuser')
        app_menu.add_item(
            'admin/groups', _('Permission groups'), url=model_url('auth.group', 'list'), order=9010, permission='is_superuser')
        app_menu.add_item('admin/logs', _('Logs'), url=model_url('trionyx.log', 'list'), order=9090, permission='is_superuser')

        # Add User renderer
        from trionyx.models import get_class
        from trionyx.renderer import renderer
        renderer.register(
            get_class('trionyx.User'),
            lambda value, **options: """<img src="{url}" class="avatar-sm" title="{user}">""".format(
                url=value.avatar.url if value.avatar else static('img/avatar.png'),
                user=str(value)
            ))

    def auto_load_app_modules(self, modules):
        """Auto load app modules"""
        for app in apps.get_app_configs():
            for module in modules:
                try:
                    import_module('{}.{}'.format(app.module.__package__, module))
                except ImportError as e:
                    if str(e) != "No module named '{}.{}'".format(app.module.__package__, module):
                        raise e

    class User(ModelConfig):
        """User config"""

        list_default_fields = ['created_at', 'email', 'first_name', 'last_name', 'is_active', 'is_superuser']
        auditlog_ignore_fields = ['last_online', 'last_login']
        verbose_name = '{email}'

    class Log(ModelConfig):
        """Log config"""

        verbose_name = '{message}'

        disable_add = True
        disable_change = True
        disable_delete = True

        global_search = False
        auditlog_disable = True
        api_disable = True

        list_default_fields = ['message', 'level', 'last_event', 'log_count']
        list_default_sort = '-last_event'
        list_fields = [
            {
                'field': 'level',
                'label': _('Level'),
                'renderer': render_level
            }
        ]

    class AuditLogEntry(ModelConfig):
        """AuditlogEntry config"""

        disable_search_index = True

        disable_add = True
        disable_change = True
        disable_delete = True

        auditlog_disable = True
        api_disable = True

    class Task(ModelConfig):
        """Task config"""

        verbose_name = '{description}'

        list_default_fields = ['created_at', 'description', 'user', 'status', 'progress', 'started_at', 'execution_time', 'result']

        list_fields = [
            {
                'field': 'status',
                'renderer': render_status,
            },
            {
                'field': 'progress',
                'renderer': render_progress,
            },
            {
                'field': 'execution_time',
                'renderer': lambda obj, *args, **options: str(timezone.timedelta(seconds=obj.execution_time)),
            },
        ]

        disable_search_index = True
        menu_exclude = True

        disable_add = True
        disable_change = True
        disable_delete = True

        auditlog_disable = True
        api_disable = True
